#!/usr/bin/env python

import sys
from typing import Any

from antlr4 import CommonTokenStream, FileStream, TerminalNode

from fandango.converters.FandangoConverter import FandangoConverter
from fandango.converters.antlr.ANTLRv4Lexer import ANTLRv4Lexer
from fandango.converters.antlr.ANTLRv4Parser import ANTLRv4Parser
from fandango.converters.antlr.ANTLRv4ParserVisitor import ANTLRv4ParserVisitor
from fandango.language.parse import PythonAntlrErrorListener


class ANTLRFandangoConverterVisitor(ANTLRv4ParserVisitor):
    def strip_quotes(self, s: str) -> str:
        """Strip quotes from a string."""
        if s.startswith('"') and s.endswith('"'):
            return s[1:-1]
        elif s.startswith("'") and s.endswith("'"):
            return s[1:-1]
        return s

    def invert_range(self, range_str: str) -> str:
        """Invert a character range string."""
        if range_str.startswith("[") and range_str.endswith("]"):
            content = range_str[1:-1]
            if content.startswith("^"):
                content = content[1:]  # Remove the negation
            else:
                content = "^" + content  # Add negation
            return f"[{content}]"

        self.addNote(f"cannot invert {range_str}")
        return "not " + range_str  # Cannot handle

    def quote(self, s: str) -> str:
        """Quote a string if it is not already quoted."""
        if "'" in s and '"' in s:
            return f'"""{s}"""'
        if "'" in s:
            return f'"{s}"'
        return f"'{s}'"

    def rquote(self, s: str) -> str:
        return "r" + self.quote(s)

    def addNote(self, message: str):
        """Add a note to the current rule"""
        if getattr(self, "notes", None) is None:
            self.notes = []
        self.notes.append(message)

    def getNotes(self) -> str:
        """Retrieve notes for current rule"""
        if getattr(self, "notes", None) is None:
            self.notes = []
        if not self.notes:
            return ""
        s = "  # NOTE: " + "; ".join(self.notes)
        self.notes = []
        return s

    def visitChildren(self, ctx: Any, sep: str = "", altEmpty: str = "") -> str:
        """Visit all children of a context `ctx`. Separate them with `sep`. If they evaluate to an empty string, use `altEmpty` instead."""
        if ctx is None or ctx.children is None:
            return ""

        children_s = []
        for child in ctx.children or []:
            s = self.visit(child)
            if s == "":
                s = altEmpty
            if s:
                children_s.append(s)
        return sep.join(children_s)

    def visitGrammarDecl(self, ctx: ANTLRv4Parser.GrammarDeclContext):
        return "# " + ctx.identifier().getText() + "\n"

    def visitParserRuleSpec(self, ctx: ANTLRv4Parser.ParserRuleSpecContext):
        if ctx.ruleModifiers():
            self.addNote(f"had modifier(s) '{ctx.ruleModifiers().getText()}'")
        if ctx.argActionBlock():
            self.addNote(f"had action block {ctx.argActionBlock().getText()}")
        if ctx.ruleReturns():
            self.addNote(f"had 'returns' clause {ctx.ruleReturns().getText()}")
        if ctx.throwsSpec():
            self.addNote(f"had 'throws' clause {ctx.throwsSpec().getText()}")
        if ctx.rulePrequel():
            self.addNote(f"had rule prequel '{ctx.rulePrequel().getText()}'")
        if ctx.exceptionGroup() and ctx.exceptionGroup().getText():
            self.addNote(f"had exception group '{ctx.exceptionGroup().getText()}'")

        nonterminal = ctx.RULE_REF().getText()
        return (
            f"<{nonterminal}> ::= " + self.visitChildren(ctx) + self.getNotes() + "\n"
        )

    def visitRuleAltList(self, ctx: ANTLRv4Parser.RuleAltListContext):
        return self.visitChildren(ctx, sep=" | ", altEmpty="''")

    def visitAltList(self, ctx: ANTLRv4Parser.AltListContext):
        return self.visitChildren(ctx, sep=" | ", altEmpty="''")

    def visitAlternative(self, ctx: ANTLRv4Parser.AlternativeContext):
        return self.visitChildren(ctx, sep=" ")

    def visitLexerAltList(self, ctx: ANTLRv4Parser.LexerAltListContext):
        return self.visitChildren(ctx, sep=" | ")

    def visitEbnf(self, ctx: ANTLRv4Parser.EbnfContext):
        if ctx.blockSuffix():
            return self.visit(ctx.block()) + self.visit(ctx.blockSuffix())
        return super().visitEbnf(ctx)

    def visitElement(self, ctx: ANTLRv4Parser.ElementContext):
        s = ""
        if ctx.labeledElement():
            s = self.visit(ctx.labeledElement())
        if ctx.atom():
            s = self.visit(ctx.atom())
        if ctx.ebnf():
            s = self.visit(ctx.ebnf())
        if ctx.actionBlock():
            s = self.visit(ctx.actionBlock())
        if ctx.ebnfSuffix():
            s += self.visit(ctx.ebnfSuffix())
        return s

    def visitWildcard(self, ctx: ANTLRv4Parser.WildcardContext):
        return self.rquote(".")

    def visitEbnfSuffix(self, ctx: ANTLRv4Parser.EbnfSuffixContext):
        suffix = ctx.getText() if ctx else ""
        if len(suffix) > 1:
            # ANTLR has non-greedy suffixes such as *? and +?
            self.addNote(f"was '{suffix}'")
            suffix = suffix[0]
        return suffix

    def visitLexerRuleSpec(self, ctx: ANTLRv4Parser.LexerRuleSpecContext):
        # This method can be customized to handle rule specifications
        nonterminal = ctx.TOKEN_REF().getText()
        return (
            f"<{nonterminal}> ::= " + self.visitChildren(ctx) + self.getNotes() + "\n"
        )

    def visitRuleref(self, ctx: ANTLRv4Parser.RulerefContext):
        rule_name = ctx.RULE_REF().getText()
        return f"<{rule_name}>"

    def visitTerminalDef(self, ctx: ANTLRv4Parser.TerminalDefContext):
        if ctx.STRING_LITERAL():
            terminal = ctx.STRING_LITERAL().getText()
            return f"{terminal}"
        elif ctx.TOKEN_REF():
            token = ctx.TOKEN_REF().getText()
            return f"<{token}>"
        # elif ctx.LEXER_CHAR_SET():
        #     char_set = ctx.LEXER_CHAR_SET().getText()
        #     return f"{char_set}"
        else:
            elem = ctx.getText()
            return f"{elem}"

    def visitLexerElement(self, ctx: ANTLRv4Parser.LexerElementContext):
        if ctx.actionBlock():
            self.visit(ctx.actionBlock())

        s = ""
        if ctx.lexerAtom():
            s = self.visit(ctx.lexerAtom())
        if ctx.lexerBlock():
            s = self.visit(ctx.lexerBlock())
        if ctx.ebnfSuffix():
            s += self.visit(ctx.ebnfSuffix())
        return s

    def visitActionBlock(self, ctx: ANTLRv4Parser.ActionBlockContext):
        self.addNote(f"was {ctx.getText()}")
        return super().visitActionBlock(ctx)

    def visitLexerCommand(self, ctx: ANTLRv4Parser.LexerCommandContext):
        self.addNote(f"was '-> {ctx.getText()}'")
        return super().visitLexerCommand(ctx)

    def visitPredicateOption(self, ctx):
        self.addNote(f"was '{ctx.getText()}'")
        return super().visitPredicateOption(ctx)

    def visitElementOption(self, ctx):
        self.addNote(f"was '{ctx.getText()}'")
        return super().visitElementOption(ctx)

    def visitLexerElements(self, ctx: ANTLRv4Parser.LexerElementsContext):
        return self.visitChildren(ctx, sep=" ")

    def visitLexerBlock(self, ctx: ANTLRv4Parser.LexerBlockContext):
        return "(" + super().visitLexerBlock(ctx) + ")"

    def visitBlockSet(self, ctx: ANTLRv4Parser.BlockSetContext):
        return "(" + super().visitBlockSet(ctx) + ")"

    def visitBlock(self, ctx: ANTLRv4Parser.BlockContext):
        if ctx.ruleAction():
            self.addNote(f"action was {ctx.ruleAction().getText()}")
        return "(" + super().visitBlock(ctx) + ")"

    def visitLexerAtom(self, ctx: ANTLRv4Parser.LexerAtomContext):
        if ctx.LEXER_CHAR_SET():
            return self.rquote(ctx.LEXER_CHAR_SET().getText())
        return super().visitLexerAtom(ctx)

    def visitCharacterRange(self, ctx: ANTLRv4Parser.CharacterRangeContext):
        range_start = self.strip_quotes(ctx.STRING_LITERAL(0).getText())
        range_end = self.strip_quotes(ctx.STRING_LITERAL(1).getText())
        return self.rquote(f"[{range_start}-{range_end}]")

    def visitNotSet(self, ctx: ANTLRv4Parser.NotSetContext):
        if ctx.setElement():
            return self.rquote(self.invert_range(ctx.setElement().getText()))
        if ctx.blockSet():
            return self.rquote(self.invert_range(ctx.blockSet().getText()))
        return super().visitChildren(ctx)

    def visitTerminal(self, node: TerminalNode):
        return super().visitTerminal(node)


class ANTLRFandangoConverter(FandangoConverter):
    """Convert ANTLR4 grammar to Fandango format."""

    def __init__(self, filename: str):
        """Initialize with given grammar file"""
        super().__init__(filename)

    def to_fan(self, **kw_args) -> str:
        """Convert the grammar spec to Fandango format"""
        # Read the grammar spec
        input_stream = FileStream(self.filename)
        error_listener = PythonAntlrErrorListener(self.filename)

        lexer = ANTLRv4Lexer(input_stream)
        lexer.removeErrorListeners()
        lexer.addErrorListener(error_listener)

        stream = CommonTokenStream(lexer)
        parser = ANTLRv4Parser(stream)
        parser.removeErrorListeners()
        parser.addErrorListener(error_listener)

        # Start parsing at the 'grammarSpec' rule
        tree = parser.grammarSpec()

        # Create a visitor and evaluate the expression
        converter = ANTLRFandangoConverterVisitor()
        spec = converter.visit(tree)

        header = f"""# Automatically generated from {self.filename!r}.
#
"""
        return header + spec


if __name__ == "__main__":
    for filename in sys.argv[1:]:
        converter = ANTLRFandangoConverter(filename)
        print(converter.to_fan(), end="")



import random
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Optional

import fandango.language.grammar.nodes as nodes
from fandango.errors import FandangoValueError
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.concatenation import Concatenation
from fandango.language.grammar.nodes.node import Node, NodeType
from fandango.language.grammar.nodes.terminal import TerminalNode
from fandango.language.symbols.terminal import Terminal
from fandango.language.tree import DerivationTree

if TYPE_CHECKING:
    import fandango


class Repetition(Node):
    def __init__(
        self,
        node: Node,
        grammar_settings: Sequence[HasSettings],
        id: str = "",
        min_: int = 0,
        max_: Optional[int] = None,
    ):
        self.id = id
        self.min = min_
        self._max = max_
        self.node = node
        self.bounds_constraint: Optional[
            "fandango.constraints.repetition_bounds.RepetitionBoundsConstraint"
        ] = None
        self.iteration = 0
        if min_ < 0:
            raise FandangoValueError(
                f"Minimum repetitions {min_} must be greater than or equal to 0"
            )
        if self.max <= 0 or self.max < min_:
            raise FandangoValueError(
                f"Maximum repetitions {self.max} must be greater than 0 or greater than min {min_}"
            )
        super().__init__(NodeType.REPETITION, grammar_settings)

    @property
    def internal_max(self):
        return self._max

    @property
    def max(self):
        if self._max is None:
            return nodes.MAX_REPETITIONS
        return self._max

    def accept(
        self,
        visitor: "fandango.language.grammar.node_visitors.node_visitor.NodeVisitor",
    ):
        return visitor.visitRepetition(self)

    def fuzz(
        self,
        parent: DerivationTree,
        grammar: "fandango.language.grammar.grammar.Grammar",
        max_nodes: int = 100,
        in_message: bool = False,
        override_current_iteration: Optional[int] = None,
        override_starting_repetition: int = 0,
        override_iterations_to_perform: Optional[int] = None,
    ):
        prev_parent_size = parent.size()
        prev_children_len = len(parent.children)
        if override_current_iteration is None:
            self.iteration += 1
            current_iteration = self.iteration
        else:
            current_iteration = override_current_iteration

        rep_goal = random.randint(self.min, self.max)
        if override_iterations_to_perform is not None:
            rep_goal = override_iterations_to_perform - override_starting_repetition

        reserved_max_nodes = self.distance_to_completion

        for rep in range(rep_goal):
            current_rep = rep + override_starting_repetition
            if self.node.distance_to_completion >= max_nodes:
                if rep > self.min and override_iterations_to_perform is None:
                    break
                self.node.fuzz(parent, grammar, 0, in_message)
            else:
                reserved_max_nodes -= self.node.distance_to_completion
                self.node.fuzz(
                    parent, grammar, int(max_nodes - reserved_max_nodes), in_message
                )
            for child in parent.children[prev_children_len:]:
                child.origin_repetitions.insert(
                    0, (self.id, current_iteration, current_rep)
                )
            max_nodes -= parent.size() - prev_parent_size
            prev_parent_size = parent.size()
            prev_children_len = len(parent.children)

    def format_as_spec(self) -> str:
        if self.min == self.max:
            return f"{self.node.format_as_spec()}{{{self.min}}}"
        return f"{self.node.format_as_spec()}{{{self.min},{self.max}}}"

    def descendents(
        self, grammar: "fandango.language.grammar.grammar.Grammar"
    ) -> Iterator["Node"]:
        base: list = []
        if self.min == 0:
            base.append(TerminalNode(Terminal(""), self._grammar_settings))
        if self.min <= 1 <= self.max:
            base.append(self.node)
        yield Alternative(
            base
            + [
                Concatenation([self.node] * r, self._grammar_settings)
                for r in range(max(2, self.min), self.max + 1)
            ],
            self._grammar_settings,
        )

    def children(self):
        return [self.node]

    def in_parties(self, parties: list[str]) -> bool:
        return self.node.in_parties(parties)


class Star(Repetition):
    def __init__(
        self,
        node: Node,
        grammar_settings: Sequence[HasSettings],
        id: str = "",
    ):
        super().__init__(node, grammar_settings, id, min_=0)

    def accept(
        self,
        visitor: "fandango.language.grammar.node_visitors.node_visitor.NodeVisitor",
    ):
        return visitor.visitStar(self)

    def format_as_spec(self) -> str:
        return self.node.format_as_spec() + "*"


class Plus(Repetition):
    def __init__(
        self,
        node: Node,
        grammar_settings: Sequence[HasSettings],
        id: str = "",
    ):
        super().__init__(node, grammar_settings, id, min_=1)

    def fuzz(
        self,
        parent: DerivationTree,
        grammar: "fandango.language.grammar.grammar.Grammar",
        max_nodes: int = 100,
        in_message: bool = False,
        override_current_iteration: Optional[int] = None,
        override_starting_repetition: int = 0,
        override_iterations_to_perform: Optional[int] = None,
    ):
        # Gmutator mutation (1b)
        if random.random() < self.settings.get("plus_should_return_nothing"):
            return  # nop, don't add a node
        else:
            return super().fuzz(
                parent,
                grammar,
                max_nodes,
                in_message,
                override_current_iteration,
                override_starting_repetition,
                override_iterations_to_perform,
            )

    def accept(
        self,
        visitor: "fandango.language.grammar.node_visitors.node_visitor.NodeVisitor",
    ):
        return visitor.visitPlus(self)

    def format_as_spec(self) -> str:
        return self.node.format_as_spec() + "+"


class Option(Repetition):
    def __init__(
        self,
        node: Node,
        grammar_settings: Sequence[HasSettings],
        id: str = "",
    ):
        super().__init__(node, grammar_settings, id, min_=0, max_=1)

    def fuzz(
        self,
        parent: DerivationTree,
        grammar: "fandango.language.grammar.grammar.Grammar",
        max_nodes: int = 100,
        in_message: bool = False,
        override_current_iteration: Optional[int] = None,
        override_starting_repetition: int = 0,
        override_iterations_to_perform: Optional[int] = None,
    ):
        # Gmutator mutation (1c)
        should_return_multiple = random.random() < self.settings.get(
            "option_should_return_multiple"
        )
        if should_return_multiple:
            repetition = Repetition(
                self.node, self._grammar_settings, id=self.id, min_=2
            )
            repetition.distance_to_completion = self.node.distance_to_completion * 2 + 1
            return repetition.fuzz(
                parent,
                grammar,
                max_nodes,
                in_message,
                override_current_iteration,
                override_starting_repetition,
                override_iterations_to_perform,
            )
        else:
            return super().fuzz(
                parent,
                grammar,
                max_nodes,
                in_message,
                override_current_iteration,
                override_starting_repetition,
                override_iterations_to_perform,
            )

    def accept(
        self,
        visitor: "fandango.language.grammar.node_visitors.node_visitor.NodeVisitor",
    ):
        return visitor.visitOption(self)

    def format_as_spec(self) -> str:
        return self.node.format_as_spec() + "?"

    def descendents(
        self, grammar: "fandango.language.grammar.grammar.Grammar"
    ) -> Iterator["Node"]:
        yield from (self.node, TerminalNode(Terminal(""), self._grammar_settings))



#!/usr/bin/env python3

import sys

from fandango.converters.FandangoConverter import FandangoConverter
from fandango.language.parse import parse_spec


class FandangoFandangoConverter(FandangoConverter):
    """Convert (normalize) Fandango grammar to Fandango format"""

    def __init__(self, filename: str, parties: list[str] | None = None):
        super().__init__(filename)
        self.parties = parties or []

    def to_fan(self, **kw_args) -> str:
        """Convert the grammar spec to Fandango format"""
        contents = open(self.filename, "r").read()
        parsed_spec = parse_spec(
            contents, filename=self.filename, use_cache=False, parties=self.parties
        )
        header = f"""# Automatically generated from {self.filename!r}.
#
"""
        return header + str(parsed_spec)


if __name__ == "__main__":
    for filename in sys.argv[1:]:
        converter = FandangoFandangoConverter(filename)
        print(converter.to_fan(), end="")



from itertools import combinations, permutations
import random
from typing import TYPE_CHECKING
from collections.abc import Iterator, Sequence
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.nodes.concatenation import Concatenation
from fandango.language.grammar.nodes.node import Node, NodeType
from fandango.language.tree import DerivationTree

if TYPE_CHECKING:
    import fandango


class Alternative(Node):
    def __init__(
        self,
        alternatives: list[Node],
        grammar_settings: Sequence[HasSettings],
        id: str = "",
    ):
        assert len(alternatives) > 0, "alternatives must be non-empty"
        self.id = id
        self.alternatives = alternatives
        super().__init__(NodeType.ALTERNATIVE, grammar_settings)

    def fuzz(
        self,
        parent: DerivationTree,
        grammar: "fandango.language.grammar.grammar.Grammar",
        max_nodes: int = 100,
        in_message: bool = False,
    ):
        in_range_nodes: Sequence[Node] = list(
            filter(lambda x: x.distance_to_completion < max_nodes, self.alternatives)
        )

        if len(in_range_nodes) == 0:
            min_ = min(self.alternatives, key=lambda x: x.distance_to_completion)
            in_range_nodes = [
                a
                for a in self.alternatives
                if a.distance_to_completion <= min_.distance_to_completion
            ]
            max_nodes = 0

        # Gmutator mutation (2)
        if random.random() < self.settings.get("alternatives_should_concatenate"):
            if len(in_range_nodes) < 2:
                pass  # can't concatenate less than 2 nodes
            else:
                concatenations: list[list[Node]] = []
                for r in range(2, len(in_range_nodes) + 1):
                    concatenations.extend(map(list, permutations(in_range_nodes, r)))
                concats = [
                    Concatenation(concatenation, self._grammar_settings)
                    for concatenation in concatenations
                ]
                for node in concats:
                    node.distance_to_completion = (
                        sum(n.distance_to_completion for n in node.nodes) + 1
                    )
                in_range_nodes = concats

        random.choice(in_range_nodes).fuzz(parent, grammar, max_nodes, in_message)

    def accept(
        self,
        visitor: "fandango.language.grammar.node_visitors.node_visitor.NodeVisitor",
    ):
        return visitor.visitAlternative(self)

    def children(self):
        return self.alternatives

    def slice_parties(self, parties: list[str]) -> None:
        self.alternatives = [
            node for node in self.alternatives if node.in_parties(parties)
        ]
        super().slice_parties(parties)

    def __getitem__(self, item):
        return self.alternatives.__getitem__(item)

    def __len__(self):
        return len(self.alternatives)

    def format_as_spec(self) -> str:
        return (
            "(" + " | ".join(map(lambda x: x.format_as_spec(), self.alternatives)) + ")"
        )

    def descendents(
        self, grammar: "fandango.language.grammar.grammar.Grammar"
    ) -> Iterator["Node"]:
        yield from self.alternatives



from typing import Optional, TypeVar, Generic
from collections.abc import Generator

# Define type variables for generator type and return type
GT = TypeVar("GT")  # Generator Type
RT = TypeVar("RT")  # Return Type


class GeneratorWithReturn(Generic[GT, RT]):
    def __init__(self, generator: Generator[GT, None, RT]):
        self.generator = generator
        self._return_value: Optional[RT] = None

    def __iter__(self):
        self._return_value = yield from self.generator

    @property
    def return_value(self) -> RT:
        """Get the return value of the generator.

        Raises:
            RuntimeError: If the generator hasn't been fully executed yet.
        """
        if self._return_value is None:
            raise RuntimeError(
                "Generator hasn't been fully executed yet. The return value is only available after complete iteration."
            )
        return self._return_value

    def collect(self) -> tuple[list[GT], RT]:
        """
        :return: A tuple containing the list of yielded values and the return value.
        """
        return list(self), self.return_value



#!/usr/bin/env pytest
from fandango import parse
from fandango.api import Fandango
from fandango.language.grammar import FuzzingMode
from fandango.language.tree import DerivationTree
from .utils import RESOURCES_ROOT


def test_msg_exchange():
    with open(RESOURCES_ROOT / "minimal_io.fan") as f:
        spec = f.read()
    fandango = Fandango(spec, use_stdlib=False, use_cache=False)
    result_list = fandango.fuzz(mode=FuzzingMode.IO, population_size=1)
    assert len(result_list) == 1
    result = result_list[0]
    assert isinstance(result, DerivationTree)
    messages = result.protocol_msgs()
    assert len(messages) == 4
    assert messages[0].sender == "Fuzzer"
    assert messages[0].recipient == "Extern"
    assert str(messages[0].msg) == "ping\n"
    assert messages[1].sender == "Extern"
    assert messages[1].recipient == "Fuzzer"
    assert str(messages[1].msg) == "pong\n"
    assert messages[2].sender == "Fuzzer"
    assert messages[2].recipient == "Extern"
    assert str(messages[2].msg) == "puff\n"
    assert messages[3].sender == "Extern"
    assert messages[3].recipient == "Fuzzer"
    assert str(messages[3].msg) == "paff\n"



from typing import cast
import unittest

from fandango import parse
from fandango.constraints.constraint import Constraint
from fandango.constraints.comparison import ComparisonConstraint
from fandango.constraints.exists import ExistsConstraint
from fandango.constraints.forall import ForallConstraint
from fandango.constraints.failing_tree import Comparison
from fandango.language.symbols import NonTerminal, Terminal
from fandango.language.tree import DerivationTree
from fandango.language.grammar.grammar import Grammar
from fandango.language.search import (
    StarSearch,
    RuleSearch,
    PopulationSearch,
    AttributeSearch,
    DescendantAttributeSearch,
)


class TestStar(unittest.TestCase):
    EXAMPLE = """
<start> ::= <a> <b> <c> <c>
<a> ::= "a" | "b"
<b> ::= "c" | "d"
<c> ::= "e" | "f" | "g" | "h"

where any(<x> == "a" for <x> in *<a>)
where all(<x> == "c" for <x> in *<b>)
where {str(x) for x in *<c>} == {"e", "f"}
"""

    VALID = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("a"))]),
            DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("c"))]),
            DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("f"))]),
            DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("e"))]),
        ],
    )

    INVALID_A = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("b"))]),
            DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("c"))]),
            DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("f"))]),
            DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("e"))]),
        ],
    )
    INVALID_B = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("a"))]),
            DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("d"))]),
            DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("f"))]),
            DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("e"))]),
        ],
    )
    INVALID_C = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("a"))]),
            DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("c"))]),
            DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("e"))]),
            DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("e"))]),
        ],
    )

    grammar: Grammar
    constraints: list[Constraint]
    any_constraint: ExistsConstraint
    all_constraint: ForallConstraint
    expression_constraint: ComparisonConstraint

    @classmethod
    def setUpClass(cls):
        # set parser to python
        # fandango.Fandango.parser = "python"
        grammar, constraints = parse(
            TestStar.EXAMPLE, use_stdlib=False, use_cache=False
        )
        for e in constraints:
            assert isinstance(e, Constraint)
        cls.constraints = cast(list[Constraint], constraints)
        assert grammar is not None
        cls.grammar = grammar
        assert isinstance(cls.constraints[0], ExistsConstraint)
        cls.any_constraint = cls.constraints[0]
        assert isinstance(cls.constraints[1], ForallConstraint)
        cls.all_constraint = cls.constraints[1]
        assert isinstance(cls.constraints[2], ComparisonConstraint)
        cls.expression_constraint = cls.constraints[2]

    def test_parse_star(self):
        self.assertIsNotNone(self.grammar)
        self.assertIsNotNone(self.constraints)
        self.assertEqual(len(self.grammar.rules), 4)
        self.assertEqual(len(self.constraints), 3)
        self.assertIn("<start>", self.grammar)
        self.assertIn("<a>", self.grammar)
        self.assertIn("<b>", self.grammar)
        self.assertIn("<c>", self.grammar)
        # Check constraints
        # Check exists constraint
        assert isinstance(self.any_constraint, ExistsConstraint)
        assert isinstance(self.any_constraint.statement, ComparisonConstraint)
        self.assertEqual(self.any_constraint.statement.operator, Comparison.EQUAL)
        tmp_var = self.any_constraint.statement.left
        self.assertIn(tmp_var, self.any_constraint.statement.searches)
        rule_search = self.any_constraint.statement.searches[tmp_var]
        assert isinstance(rule_search, RuleSearch)
        self.assertEqual(rule_search.symbol, NonTerminal("<x>"))
        self.assertEqual(eval(self.any_constraint.statement.right), "a")
        self.assertEqual(self.any_constraint.bound, NonTerminal("<x>"))
        assert isinstance(self.any_constraint.search, StarSearch)
        assert isinstance(self.any_constraint.search.base, RuleSearch)
        self.assertEqual(self.any_constraint.search.base.symbol, NonTerminal("<a>"))

        # Check forall constraint
        assert isinstance(self.all_constraint, ForallConstraint)
        assert isinstance(self.all_constraint.statement, ComparisonConstraint)
        self.assertEqual(self.all_constraint.statement.operator, Comparison.EQUAL)
        tmp_var = self.all_constraint.statement.left
        self.assertIn(tmp_var, self.all_constraint.statement.searches)
        rule_search = self.all_constraint.statement.searches[tmp_var]
        assert isinstance(rule_search, RuleSearch)
        self.assertEqual(rule_search.symbol, NonTerminal("<x>"))
        self.assertEqual(eval(self.all_constraint.statement.right), "c")
        self.assertEqual(self.all_constraint.bound, NonTerminal("<x>"))
        assert isinstance(self.all_constraint.search, StarSearch)
        assert isinstance(self.all_constraint.search.base, RuleSearch)
        self.assertEqual(self.all_constraint.search.base.symbol, NonTerminal("<b>"))

        # Check expression constraint
        assert isinstance(self.expression_constraint, ComparisonConstraint)
        self.assertEqual(self.expression_constraint.operator, Comparison.EQUAL)
        tmp_var = self.expression_constraint.left
        self.assertTrue(tmp_var.startswith("{str(x) for x in "))
        self.assertTrue(tmp_var.endswith("}"))
        tmp_var = tmp_var[17:-1]  # Remove the prefix and suffix
        self.assertIn(tmp_var, self.expression_constraint.searches)
        search = self.expression_constraint.searches[tmp_var]
        assert isinstance(search, StarSearch)
        assert isinstance(search.base, RuleSearch)
        self.assertEqual(search.base.symbol, NonTerminal("<c>"))
        self.assertEqual(eval(self.expression_constraint.right), {"e", "f"})

    def test_star_constraint_valid(self):
        for constraint in self.constraints:
            self.assertTrue(constraint.check(self.VALID), constraint)

    def test_invalid_a(self):
        self.assertFalse(
            self.any_constraint.check(self.INVALID_A),
            "Invalid <a> should not satisfy the exists constraint",
        )
        self.assertTrue(
            self.all_constraint.check(self.INVALID_A),
            "Invalid <a> should satisfy the forall constraint",
        )
        self.assertTrue(
            self.expression_constraint.check(self.INVALID_A),
            "Invalid <a> should satisfy the expression constraint",
        )

    def test_invalid_b(self):
        self.assertTrue(
            self.any_constraint.check(self.INVALID_B),
            "Invalid <b> should satisfy the exists constraint",
        )
        self.assertFalse(
            self.all_constraint.check(self.INVALID_B),
            "Invalid <b> should not satisfy the forall constraint",
        )
        self.assertTrue(
            self.expression_constraint.check(self.INVALID_B),
            "Invalid <b> should satisfy the expression constraint",
        )

    def test_invalid_c(self):
        self.assertTrue(
            self.any_constraint.check(self.INVALID_C),
            "Invalid <c> should satisfy the exists constraint",
        )
        self.assertTrue(
            self.all_constraint.check(self.INVALID_C),
            "Invalid <c> should satisfy the forall constraint",
        )
        self.assertFalse(
            self.expression_constraint.check(self.INVALID_C),
            "Invalid <c> should not satisfy the expression constraint",
        )


class TestPopulation(unittest.TestCase):
    EXAMPLE = """
<start> ::= <a> <b> <c>
<a> ::= "a" | "b"
<b> ::= "c" | "d"
<c> ::= "e" | "f" | "g" | "h"

where any(<x> == "a" for <x> in **<a>)
where all(<x> == "c" for <x> in **<b>)
where {str(x) for x in **<c>} == {"e", "f"}
"""

    VALID_POPULATION = [
        DerivationTree(
            NonTerminal("<start>"),
            [
                DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("a"))]),
                DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("c"))]),
                DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("f"))]),
            ],
        ),
        DerivationTree(
            NonTerminal("<start>"),
            [
                DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("b"))]),
                DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("c"))]),
                DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("e"))]),
            ],
        ),
    ]

    INVALID_POPULATION_A = [
        DerivationTree(
            NonTerminal("<start>"),
            [
                DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("b"))]),
                DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("c"))]),
                DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("f"))]),
            ],
        ),
        DerivationTree(
            NonTerminal("<start>"),
            [
                DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("b"))]),
                DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("c"))]),
                DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("e"))]),
            ],
        ),
    ]

    INVALID_POPULATION_B = [
        DerivationTree(
            NonTerminal("<start>"),
            [
                DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("a"))]),
                DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("d"))]),
                DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("f"))]),
            ],
        ),
        DerivationTree(
            NonTerminal("<start>"),
            [
                DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("b"))]),
                DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("c"))]),
                DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("e"))]),
            ],
        ),
    ]

    INVALID_POPULATION_C = [
        DerivationTree(
            NonTerminal("<start>"),
            [
                DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("a"))]),
                DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("c"))]),
                DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("e"))]),
            ],
        ),
        DerivationTree(
            NonTerminal("<start>"),
            [
                DerivationTree(NonTerminal("<a>"), [DerivationTree(Terminal("b"))]),
                DerivationTree(NonTerminal("<b>"), [DerivationTree(Terminal("c"))]),
                DerivationTree(NonTerminal("<c>"), [DerivationTree(Terminal("g"))]),
            ],
        ),
    ]

    grammar: Grammar
    constraints: list[Constraint]
    any_constraint: ExistsConstraint
    all_constraint: ForallConstraint
    expression_constraint: ComparisonConstraint

    @classmethod
    def setUpClass(cls):
        grammar, constraints = parse(
            TestPopulation.EXAMPLE, use_stdlib=False, use_cache=False
        )
        for e in constraints:
            assert isinstance(e, Constraint)
        cls.constraints = cast(list[Constraint], constraints)

        assert grammar is not None
        cls.grammar = grammar
        assert isinstance(cls.constraints[0], ExistsConstraint)
        cls.any_constraint = cls.constraints[0]
        assert isinstance(cls.constraints[1], ForallConstraint)
        cls.all_constraint = cls.constraints[1]
        assert isinstance(cls.constraints[2], ComparisonConstraint)
        cls.expression_constraint = cls.constraints[2]

    def test_parse_population_star(self):
        self.assertIsNotNone(self.grammar)
        self.assertIsNotNone(self.constraints)
        self.assertEqual(len(self.grammar.rules), 4)
        self.assertEqual(len(self.constraints), 3)
        self.assertIn("<start>", self.grammar)
        self.assertIn("<a>", self.grammar)
        self.assertIn("<b>", self.grammar)
        self.assertIn("<c>", self.grammar)
        # Check constraints
        # Check exists constraint
        assert isinstance(self.any_constraint, ExistsConstraint)
        assert isinstance(self.any_constraint.statement, ComparisonConstraint)
        self.assertEqual(self.any_constraint.statement.operator, Comparison.EQUAL)
        tmp_var = self.any_constraint.statement.left
        self.assertIn(tmp_var, self.any_constraint.statement.searches)
        rule_search = self.any_constraint.statement.searches[tmp_var]
        assert isinstance(rule_search, RuleSearch)
        self.assertEqual(rule_search.symbol, NonTerminal("<x>"))
        self.assertEqual(eval(self.any_constraint.statement.right), "a")
        self.assertEqual(self.any_constraint.bound, NonTerminal("<x>"))
        assert isinstance(self.any_constraint.search, PopulationSearch)
        assert isinstance(self.any_constraint.search.base, RuleSearch)
        self.assertEqual(self.any_constraint.search.base.symbol, NonTerminal("<a>"))

        # Check forall constraint
        assert isinstance(self.all_constraint, ForallConstraint)
        assert isinstance(self.all_constraint.statement, ComparisonConstraint)
        self.assertEqual(self.all_constraint.statement.operator, Comparison.EQUAL)
        tmp_var = self.all_constraint.statement.left
        self.assertIn(tmp_var, self.all_constraint.statement.searches)
        rule_search = self.all_constraint.statement.searches[tmp_var]
        assert isinstance(rule_search, RuleSearch)
        self.assertEqual(rule_search.symbol, NonTerminal("<x>"))
        self.assertEqual(eval(self.all_constraint.statement.right), "c")
        self.assertEqual(self.all_constraint.bound, NonTerminal("<x>"))
        assert isinstance(self.all_constraint.search, PopulationSearch)
        assert isinstance(self.all_constraint.search.base, RuleSearch)
        self.assertEqual(self.all_constraint.search.base.symbol, NonTerminal("<b>"))

        # Check expression constraint
        assert isinstance(self.expression_constraint, ComparisonConstraint)
        self.assertEqual(self.expression_constraint.operator, Comparison.EQUAL)
        tmp_var = self.expression_constraint.left
        self.assertTrue(tmp_var.startswith("{str(x) for x in "))
        self.assertTrue(tmp_var.endswith("}"))
        tmp_var = tmp_var[17:-1]  # Remove the prefix and suffix
        self.assertIn(tmp_var, self.expression_constraint.searches)
        search = self.expression_constraint.searches[tmp_var]
        assert isinstance(search, PopulationSearch)
        assert isinstance(search.base, RuleSearch)
        self.assertEqual(search.base.symbol, NonTerminal("<c>"))
        self.assertEqual(eval(self.expression_constraint.right), {"e", "f"})

    def test_population_star_constraint_valid(self):
        for constraint in self.constraints:
            for tree in self.VALID_POPULATION:
                self.assertTrue(
                    constraint.check(tree, population=self.VALID_POPULATION), constraint
                )

    def test_invalid_population_a(self):
        for tree in self.INVALID_POPULATION_A:
            self.assertFalse(
                self.any_constraint.check(tree, population=self.INVALID_POPULATION_A),
                "Invalid <a> should not satisfy the exists constraint",
            )
            self.assertTrue(
                self.all_constraint.check(tree, population=self.INVALID_POPULATION_A),
                "Invalid <a> should satisfy the forall constraint",
            )
            self.assertTrue(
                self.expression_constraint.check(
                    tree, population=self.INVALID_POPULATION_A
                ),
                "Invalid <a> should satisfy the expression constraint",
            )

    def test_invalid_population_b(self):
        for tree in self.INVALID_POPULATION_B:
            self.assertTrue(
                self.any_constraint.check(tree, population=self.INVALID_POPULATION_B),
                "Invalid <b> should satisfy the exists constraint",
            )
            self.assertFalse(
                self.all_constraint.check(tree, population=self.INVALID_POPULATION_B),
                "Invalid <b> should not satisfy the forall constraint",
            )
            self.assertTrue(
                self.expression_constraint.check(
                    tree, population=self.INVALID_POPULATION_B
                ),
                "Invalid <b> should satisfy the expression constraint",
            )

    def test_invalid_population_c(self):
        for tree in self.INVALID_POPULATION_C:
            self.assertTrue(
                self.any_constraint.check(tree, population=self.INVALID_POPULATION_C),
                "Invalid <c> should satisfy the exists constraint",
            )
            self.assertTrue(
                self.all_constraint.check(tree, population=self.INVALID_POPULATION_C),
                "Invalid <c> should satisfy the forall constraint",
            )
            self.assertFalse(
                self.expression_constraint.check(
                    tree, population=self.INVALID_POPULATION_C
                ),
                "Invalid <c> should not satisfy the expression constraint",
            )


class TestStarIdentifier(unittest.TestCase):
    EXAMPLE = """
<start> ::= <a> <b> <c> <c>
<a> ::= "a" | "b"
<b> ::= "c" | "d"
<c> ::= "e" | "f" | "g" | "h"

where any(x == "a" for x in *<a>)
where all(x == "c" for x in *<b>)
where {str(x) for x in *<c>} == {"e", "f"}
"""

    grammar: Grammar
    constraints: list[Constraint]
    any_constraint: ExistsConstraint
    all_constraint: ForallConstraint
    expression_constraint: ComparisonConstraint

    @classmethod
    def setUpClass(cls):
        # set parser to python
        # fandango.Fandango.parser = "python"
        grammar, constraints = parse(
            TestStarIdentifier.EXAMPLE, use_stdlib=False, use_cache=False
        )
        for e in constraints:
            assert isinstance(e, Constraint)
        cls.constraints = cast(list[Constraint], constraints)

        assert grammar is not None
        cls.grammar = grammar

        assert isinstance(cls.constraints[0], ExistsConstraint)
        cls.any_constraint = cls.constraints[0]
        assert isinstance(cls.constraints[1], ForallConstraint)
        cls.all_constraint = cls.constraints[1]
        assert isinstance(cls.constraints[2], ComparisonConstraint)
        cls.expression_constraint = cls.constraints[2]

    def test_parse_star(self):
        self.assertIsNotNone(self.grammar)
        self.assertIsNotNone(self.constraints)
        self.assertEqual(len(self.grammar.rules), 4)
        self.assertEqual(len(self.constraints), 3)
        self.assertIn("<start>", self.grammar)
        self.assertIn("<a>", self.grammar)
        self.assertIn("<b>", self.grammar)
        self.assertIn("<c>", self.grammar)
        # Check constraints
        # Check exists constraint
        assert isinstance(self.any_constraint, ExistsConstraint)
        assert isinstance(self.any_constraint.statement, ComparisonConstraint)
        self.assertEqual(self.any_constraint.statement.operator, Comparison.EQUAL)
        tmp_var = self.any_constraint.statement.left
        self.assertNotIn(tmp_var, self.any_constraint.statement.searches)
        self.assertEqual(eval(self.any_constraint.statement.right), "a")
        self.assertEqual(self.any_constraint.bound, "x")
        self.assertEqual(tmp_var, self.any_constraint.bound)
        assert isinstance(self.any_constraint.search, StarSearch)
        assert isinstance(self.any_constraint.search.base, RuleSearch)
        self.assertEqual(self.any_constraint.search.base.symbol, NonTerminal("<a>"))

        # Check forall constraint
        assert isinstance(self.all_constraint, ForallConstraint)
        assert isinstance(self.all_constraint.statement, ComparisonConstraint)
        self.assertEqual(self.all_constraint.statement.operator, Comparison.EQUAL)
        tmp_var = self.all_constraint.statement.left
        self.assertNotIn(tmp_var, self.all_constraint.statement.searches)
        self.assertEqual(eval(self.all_constraint.statement.right), "c")
        self.assertEqual(self.all_constraint.bound, "x")
        self.assertEqual(tmp_var, self.all_constraint.bound)
        assert isinstance(self.all_constraint.search, StarSearch)
        assert isinstance(self.all_constraint.search.base, RuleSearch)
        self.assertEqual(self.all_constraint.search.base.symbol, NonTerminal("<b>"))

        # Check expression constraint
        assert isinstance(self.expression_constraint, ComparisonConstraint)
        self.assertEqual(self.expression_constraint.operator, Comparison.EQUAL)
        tmp_var = self.expression_constraint.left
        self.assertTrue(tmp_var.startswith("{str(x) for x in "))
        self.assertTrue(tmp_var.endswith("}"))
        tmp_var = tmp_var[17:-1]  # Remove the prefix and suffix
        self.assertIn(tmp_var, self.expression_constraint.searches)
        search = self.expression_constraint.searches[tmp_var]
        assert isinstance(search, StarSearch)
        assert isinstance(search.base, RuleSearch)
        self.assertEqual(search.base.symbol, NonTerminal("<c>"))
        self.assertEqual(eval(self.expression_constraint.right), {"e", "f"})

    def test_star_constraint_valid(self):
        for constraint in self.constraints:
            self.assertTrue(constraint.check(TestStar.VALID), constraint)

    def test_invalid_a(self):
        self.assertFalse(
            self.any_constraint.check(TestStar.INVALID_A),
            "Invalid <a> should not satisfy the exists constraint",
        )
        self.assertTrue(
            self.all_constraint.check(TestStar.INVALID_A),
            "Invalid <a> should satisfy the forall constraint",
        )
        self.assertTrue(
            self.expression_constraint.check(TestStar.INVALID_A),
            "Invalid <a> should satisfy the expression constraint",
        )

    def test_invalid_b(self):
        self.assertTrue(
            self.any_constraint.check(TestStar.INVALID_B),
            "Invalid <b> should satisfy the exists constraint",
        )
        self.assertFalse(
            self.all_constraint.check(TestStar.INVALID_B),
            "Invalid <b> should not satisfy the forall constraint",
        )
        self.assertTrue(
            self.expression_constraint.check(TestStar.INVALID_B),
            "Invalid <b> should satisfy the expression constraint",
        )

    def test_invalid_c(self):
        self.assertTrue(
            self.any_constraint.check(TestStar.INVALID_C),
            "Invalid <c> should satisfy the exists constraint",
        )
        self.assertTrue(
            self.all_constraint.check(TestStar.INVALID_C),
            "Invalid <c> should satisfy the forall constraint",
        )
        self.assertFalse(
            self.expression_constraint.check(TestStar.INVALID_C),
            "Invalid <c> should not satisfy the expression constraint",
        )


class TestPopulationIdentifier(unittest.TestCase):
    EXAMPLE = """
<start> ::= <a> <b> <c>
<a> ::= "a" | "b"
<b> ::= "c" | "d"
<c> ::= "e" | "f" | "g" | "h"

where any(x == "a" for x in **<a>)
where all(x == "c" for x in **<b>)
where {str(x) for x in **<c>} == {"e", "f"}
"""
    any_constraint: ExistsConstraint
    all_constraint: ForallConstraint
    expression_constraint: ComparisonConstraint
    grammar: Grammar
    constraints: list[Constraint]

    @classmethod
    def setUpClass(cls):
        grammar, constraints = parse(cls.EXAMPLE, use_stdlib=False, use_cache=False)
        for e in constraints:
            assert isinstance(e, Constraint)
        cls.constraints = cast(list[Constraint], constraints)
        assert grammar is not None
        cls.grammar = grammar

        assert len(cls.constraints) == 3
        assert isinstance(cls.constraints[0], ExistsConstraint)
        assert isinstance(cls.constraints[1], ForallConstraint)
        assert isinstance(cls.constraints[2], ComparisonConstraint)
        cls.any_constraint = cls.constraints[0]
        cls.all_constraint = cls.constraints[1]
        cls.expression_constraint = cls.constraints[2]

    def test_parse_population_star(self):
        self.assertIsNotNone(self.grammar)
        self.assertIsNotNone(self.constraints)
        self.assertEqual(len(self.grammar.rules), 4)
        self.assertEqual(len(self.constraints), 3)
        self.assertIn("<start>", self.grammar)
        self.assertIn("<a>", self.grammar)
        self.assertIn("<b>", self.grammar)
        self.assertIn("<c>", self.grammar)
        # Check constraints
        # Check exists constraint
        assert isinstance(self.any_constraint, ExistsConstraint)
        assert isinstance(self.any_constraint.statement, ComparisonConstraint)
        self.assertEqual(self.any_constraint.statement.operator, Comparison.EQUAL)
        tmp_var = self.any_constraint.statement.left
        self.assertNotIn(tmp_var, self.any_constraint.statement.searches)
        self.assertEqual(eval(self.any_constraint.statement.right), "a")
        self.assertEqual(self.any_constraint.bound, "x")
        self.assertEqual(tmp_var, self.any_constraint.bound)
        assert isinstance(self.any_constraint.search, PopulationSearch)
        assert isinstance(self.any_constraint.search.base, RuleSearch)
        self.assertEqual(self.any_constraint.search.base.symbol, NonTerminal("<a>"))

        # Check forall constraint
        assert isinstance(self.all_constraint, ForallConstraint)
        assert isinstance(self.all_constraint.statement, ComparisonConstraint)
        self.assertEqual(self.all_constraint.statement.operator, Comparison.EQUAL)
        tmp_var = self.all_constraint.statement.left
        self.assertNotIn(tmp_var, self.all_constraint.statement.searches)
        self.assertEqual(eval(self.all_constraint.statement.right), "c")
        self.assertEqual(self.all_constraint.bound, "x")
        self.assertEqual(tmp_var, self.all_constraint.bound)
        assert isinstance(self.all_constraint.search, PopulationSearch)
        assert isinstance(self.all_constraint.search.base, RuleSearch)
        self.assertEqual(self.all_constraint.search.base.symbol, NonTerminal("<b>"))

        # Check expression constraint
        assert isinstance(self.expression_constraint, ComparisonConstraint)
        self.assertEqual(self.expression_constraint.operator, Comparison.EQUAL)
        tmp_var = self.expression_constraint.left
        self.assertTrue(tmp_var.startswith("{str(x) for x in "))
        self.assertTrue(tmp_var.endswith("}"))
        tmp_var = tmp_var[17:-1]  # Remove the prefix and suffix
        self.assertIn(tmp_var, self.expression_constraint.searches)
        search = self.expression_constraint.searches[tmp_var]
        assert isinstance(search, PopulationSearch)
        assert isinstance(search.base, RuleSearch)
        self.assertEqual(search.base.symbol, NonTerminal("<c>"))
        self.assertEqual(eval(self.expression_constraint.right), {"e", "f"})

    def test_population_star_constraint_valid(self):
        for constraint in self.constraints:
            for tree in TestPopulation.VALID_POPULATION:
                self.assertTrue(
                    constraint.check(tree, population=TestPopulation.VALID_POPULATION),
                    constraint,
                )

    def test_invalid_population_a(self):
        for tree in TestPopulation.INVALID_POPULATION_A:
            self.assertFalse(
                self.any_constraint.check(
                    tree, population=TestPopulation.INVALID_POPULATION_A
                ),
                "Invalid <a> should not satisfy the exists constraint",
            )
            self.assertTrue(
                self.all_constraint.check(
                    tree, population=TestPopulation.INVALID_POPULATION_A
                ),
                "Invalid <a> should satisfy the forall constraint",
            )
            self.assertTrue(
                self.expression_constraint.check(
                    tree, population=TestPopulation.INVALID_POPULATION_A
                ),
                "Invalid <a> should satisfy the expression constraint",
            )

    def test_invalid_population_b(self):
        for tree in TestPopulation.INVALID_POPULATION_B:
            self.assertTrue(
                self.any_constraint.check(
                    tree, population=TestPopulation.INVALID_POPULATION_B
                ),
                "Invalid <b> should satisfy the exists constraint",
            )
            self.assertFalse(
                self.all_constraint.check(
                    tree, population=TestPopulation.INVALID_POPULATION_B
                ),
                "Invalid <b> should not satisfy the forall constraint",
            )
            self.assertTrue(
                self.expression_constraint.check(
                    tree, population=TestPopulation.INVALID_POPULATION_B
                ),
                "Invalid <b> should satisfy the expression constraint",
            )

    def test_invalid_population_c(self):
        for tree in TestPopulation.INVALID_POPULATION_C:
            self.assertTrue(
                self.any_constraint.check(
                    tree, population=TestPopulation.INVALID_POPULATION_C
                ),
                "Invalid <c> should satisfy the exists constraint",
            )
            self.assertTrue(
                self.all_constraint.check(
                    tree, population=TestPopulation.INVALID_POPULATION_C
                ),
                "Invalid <c> should satisfy the forall constraint",
            )
            self.assertFalse(
                self.expression_constraint.check(
                    tree, population=TestPopulation.INVALID_POPULATION_C
                ),
                "Invalid <c> should not satisfy the expression constraint",
            )


class TestStarInCombination(unittest.TestCase):
    EXAMPLE = "\n".join(
        [
            "<start> ::= <a>",
            "<a> ::= <b> | <c>",
            "<b> ::= <d> | <e>",
            "<c> ::= <e> <d>",
            '<d> ::= "d"',
            '<e> ::= "e"',
        ]
    )

    CONSTRAINT_DOT = "where all(str(x) == 'd' for x in *<a>.<b>)"
    CONSTRAINT_DOT_DOT = "where all(str(x) == 'd' for x in *<start>..<b>)"
    CONSTRAINT_POPULATION_DOT = "where all(str(x) == 'd' for x in **<a>.<b>)"
    CONSTRAINT_POPULATION_DOT_DOT = "where all(str(x) == 'd' for x in **<start>..<b>)"

    EXAMPLE_1 = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(
                NonTerminal("<a>"),
                [
                    DerivationTree(
                        NonTerminal("<b>"),
                        [
                            DerivationTree(
                                NonTerminal("<d>"),
                                [
                                    DerivationTree(Terminal("d")),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )

    EXAMPLE_2 = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(
                NonTerminal("<a>"),
                [
                    DerivationTree(
                        NonTerminal("<b>"),
                        [
                            DerivationTree(
                                NonTerminal("<e>"),
                                [
                                    DerivationTree(
                                        Terminal("e"),
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )

    EXAMPLE_3 = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(
                NonTerminal("<a>"),
                [
                    DerivationTree(
                        NonTerminal("<c>"),
                        [
                            DerivationTree(
                                NonTerminal("<e>"),
                                [
                                    DerivationTree(
                                        Terminal("e"),
                                    )
                                ],
                            ),
                            DerivationTree(
                                NonTerminal("<d>"),
                                [
                                    DerivationTree(
                                        Terminal("d"),
                                    )
                                ],
                            ),
                        ],
                    )
                ],
            )
        ],
    )

    grammar: Grammar

    @classmethod
    def setUpClass(cls):
        grammar, _ = parse(cls.EXAMPLE, use_cache=False, use_stdlib=False)
        assert grammar is not None
        cls.grammar = grammar

    def test_dot(self):
        _, constraints = parse(self.CONSTRAINT_DOT, use_cache=False, use_stdlib=False)
        self.assertEqual(len(constraints), 1)
        constraint = constraints[0]
        assert isinstance(constraint, ForallConstraint)
        self.assertEqual(constraint.bound, "x")
        statement = constraint.statement
        assert isinstance(statement, ComparisonConstraint)
        self.assertEqual(statement.left, "str(x)")
        self.assertEqual(eval(statement.right), "d")
        star = constraint.search
        assert isinstance(star, StarSearch)
        base = star.base
        assert isinstance(base, AttributeSearch)
        parent = base.base
        assert isinstance(parent, RuleSearch)
        self.assertEqual(parent.symbol, NonTerminal("<a>"))
        attribute = base.attribute
        assert isinstance(attribute, RuleSearch)
        self.assertEqual(attribute.symbol, NonTerminal("<b>"))

        self.assertTrue(constraint.check(self.EXAMPLE_1))
        self.assertFalse(constraint.check(self.EXAMPLE_2))
        self.assertTrue(constraint.check(self.EXAMPLE_3))

    def test_dot_dot(self):
        _, constraints = parse(
            self.CONSTRAINT_DOT_DOT, use_cache=False, use_stdlib=False
        )
        self.assertEqual(len(constraints), 1)
        constraint = constraints[0]
        assert isinstance(constraint, ForallConstraint)
        self.assertEqual(constraint.bound, "x")
        statement = constraint.statement
        assert isinstance(statement, ComparisonConstraint)
        self.assertEqual(statement.left, "str(x)")
        self.assertEqual(eval(statement.right), "d")
        star = constraint.search
        assert isinstance(star, StarSearch)
        base = star.base
        assert isinstance(base, DescendantAttributeSearch)
        parent = base.base
        assert isinstance(parent, RuleSearch)
        self.assertEqual(parent.symbol, NonTerminal("<start>"))
        attribute = base.attribute
        assert isinstance(attribute, RuleSearch)
        self.assertEqual(attribute.symbol, NonTerminal("<b>"))

        self.assertTrue(constraint.check(self.EXAMPLE_1))
        self.assertFalse(constraint.check(self.EXAMPLE_2))
        self.assertTrue(constraint.check(self.EXAMPLE_3))

    def test_population_dot(self):
        _, constraints = parse(
            self.CONSTRAINT_POPULATION_DOT, use_cache=False, use_stdlib=False
        )
        self.assertEqual(len(constraints), 1)
        constraint = constraints[0]
        assert isinstance(constraint, ForallConstraint)
        self.assertEqual(constraint.bound, "x")
        statement = constraint.statement
        assert isinstance(statement, ComparisonConstraint)
        self.assertEqual(statement.left, "str(x)")
        self.assertEqual(eval(statement.right), "d")
        star = constraint.search
        assert isinstance(star, PopulationSearch)
        base = star.base
        assert isinstance(base, AttributeSearch)
        parent = base.base
        assert isinstance(parent, RuleSearch)
        self.assertEqual(parent.symbol, NonTerminal("<a>"))
        attribute = base.attribute
        assert isinstance(attribute, RuleSearch)
        self.assertEqual(attribute.symbol, NonTerminal("<b>"))

        self.assertFalse(
            constraint.check(
                self.EXAMPLE_1,
                population=[self.EXAMPLE_1, self.EXAMPLE_2, self.EXAMPLE_3],
            )
        )
        self.assertTrue(
            constraint.check(
                self.EXAMPLE_1,
                population=[self.EXAMPLE_1, self.EXAMPLE_1, self.EXAMPLE_3],
            )
        )

    def test_population_dot_dot(self):
        _, constraints = parse(
            self.CONSTRAINT_POPULATION_DOT_DOT, use_cache=False, use_stdlib=False
        )
        self.assertEqual(len(constraints), 1)
        constraint = constraints[0]
        assert isinstance(constraint, ForallConstraint)
        self.assertEqual(constraint.bound, "x")
        statement = constraint.statement
        assert isinstance(statement, ComparisonConstraint)
        self.assertEqual(statement.left, "str(x)")
        self.assertEqual(eval(statement.right), "d")
        star = constraint.search
        assert isinstance(star, PopulationSearch)
        base = star.base
        assert isinstance(base, DescendantAttributeSearch)
        parent = base.base
        assert isinstance(parent, RuleSearch)
        self.assertEqual(parent.symbol, NonTerminal("<start>"))
        attribute = base.attribute
        assert isinstance(attribute, RuleSearch)
        self.assertEqual(attribute.symbol, NonTerminal("<b>"))

        self.assertFalse(
            constraint.check(
                self.EXAMPLE_1,
                population=[self.EXAMPLE_1, self.EXAMPLE_2, self.EXAMPLE_3],
            )
        )
        self.assertTrue(
            constraint.check(
                self.EXAMPLE_1, population=[self.EXAMPLE_1, self.EXAMPLE_3]
            )
        )



from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.node_visitors.node_visitor import NodeVisitor
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.concatenation import Concatenation
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.grammar.nodes.repetition import Repetition, Star, Plus, Option
from fandango.language.grammar.nodes.terminal import TerminalNode


class PacketTruncator(NodeVisitor):
    def __init__(self, grammar: "Grammar", keep_parties: set[str]):
        self.grammar = grammar
        self.keep_msg_parties = keep_parties

    def visitStar(self, node: Star):
        return self.visitRepetition(node)

    def visitPlus(self, node: Plus):
        return self.visitRepetition(node)

    def visitOption(self, node: Option):
        return self.visitRepetition(node)

    def visitAlternative(self, node: Alternative):
        for child in list(node.children()):
            if self.visit(child):
                node.alternatives.remove(child)
        if len(node.alternatives) == 0:
            return True
        return False

    def visitRepetition(self, node: Repetition):
        for child in list(node.children()):
            if self.visit(child):
                return True
        return False

    def visitConcatenation(self, node: Concatenation):
        for child in list(node.children()):
            if self.visit(child):
                node.nodes.remove(child)
        if len(node.nodes) == 0:
            return True
        return False

    def visitNonTerminalNode(self, node: NonTerminalNode):
        if node.sender is None or node.recipient is None:
            return False
        truncatable = True
        if node.sender in self.keep_msg_parties:
            truncatable = False
        if node.recipient in self.keep_msg_parties:
            truncatable = False
        return truncatable

    def visitTerminalNode(self, node: TerminalNode):
        return False



import argparse
import os
import sys
import textwrap

import fandango


def terminal_link(url: str, text: str | None = None) -> str:
    """Output URL as a link"""
    if text is None:
        text = url
    # https://gist.github.com/egmontkob/eb114294efbcd5adb1944c9f3cb5feda
    return f"\x1b]8;;{url}\x1b\\{text}\x1b]8;;\x1b\\"


def homepage_as_link() -> str:
    """Return the Fandango homepage, formatted for terminals"""
    homepage = fandango.homepage()
    if os.getenv("JUPYTER_BOOK") is not None:
        return homepage  # Don't link in Jupyter Book

    if homepage.startswith("http") and sys.stdout.isatty():
        return terminal_link(homepage)
    else:
        return homepage


def get_parser(in_command_line: bool = True) -> argparse.ArgumentParser:
    main_parser = _get_main_parser(in_command_line)

    commands = main_parser.add_subparsers(
        title="commands",
        help="The command to execute.",
        dest="command",
    )

    algorithm_parser = _get_algorithm_parser()
    settings_parser = _get_settings_parser(in_command_line)
    file_parser = _get_file_parser()
    parties_parser = _get_parties_parser()

    _populate_fuzz_parser(
        parser=commands.add_parser(
            "fuzz",
            help="Produce outputs from .fan files and test programs.",
            parents=[file_parser, algorithm_parser, settings_parser, parties_parser],
        )
    )

    _populate_parse_parser(
        parser=commands.add_parser(
            "parse",
            help="Parse input file(s) according to .fan spec.",
            parents=[file_parser, settings_parser, parties_parser],
        )
    )

    _populate_talk_parser(
        parser=commands.add_parser(
            "talk",
            help="Interact with programs, clients, and servers.",
            parents=[file_parser, algorithm_parser, settings_parser],
        )
    )

    _populate_convert_parser(
        parser=commands.add_parser(
            "convert",
            help="Convert given external spec to .fan format.",
            parents=[parties_parser],
        )
    )

    _populate_clear_cache_parser(
        parser=commands.add_parser(
            "clear-cache", help="Clear the Fandango parsing cache."
        )
    )

    if in_command_line:
        commands.add_parser("shell", help="Run an interactive shell (default).")
    else:
        _populate_shell_parser_not_in_command_line(
            parser=commands.add_parser("!", help="Execute shell command."),
        )

        commands.add_parser(
            "set",
            help="Set or print default arguments.",
            parents=[file_parser, algorithm_parser, settings_parser, parties_parser],
        )
        commands.add_parser("reset", help="Reset defaults.")

        _populate_cd_parser(parser=commands.add_parser("cd", help="Change directory."))

        commands.add_parser("exit", help="Exit Fandango.")

        _populate_python_parser(
            parser=commands.add_parser("/", help="Execute Python command.")
        )

    _populate_help_parser(
        parser=commands.add_parser("help", help="Show this help and exit.")
    )

    commands.add_parser("copyright", help="Show copyright.")

    commands.add_parser("version", help="Show version.")

    return main_parser


def _get_main_parser(in_command_line: bool) -> argparse.ArgumentParser:
    if in_command_line:
        prog = "fandango"
        epilog = textwrap.dedent(
            """\
            Use `%(prog)s help` to get a list of commands.
            Use `%(prog)s help COMMAND` to learn more about COMMAND."""
        )
    else:
        prog = ""
        epilog = textwrap.dedent(
            """\
            Use `help` to get a list of commands.
            Use `help COMMAND` to learn more about COMMAND.
            Use TAB to complete commands."""
        )
    epilog += f"\nSee {homepage_as_link()} for more information."

    main_parser = argparse.ArgumentParser(
        prog=prog,
        description="The access point to the Fandango framework",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=in_command_line,
        epilog=textwrap.dedent(epilog),
    )

    if in_command_line:
        main_parser.add_argument(
            "--version",
            action="version",
            version=f"Fandango {fandango.version()}",
            help="Show version number.",
        )

        verbosity_option = main_parser.add_mutually_exclusive_group()
        verbosity_option.add_argument(
            "--verbose",
            "-v",
            dest="verbose",
            action="count",
            help="Increase verbosity. Can be given multiple times (-vv).",
        )
        verbosity_option.add_argument(
            "--quiet",
            "-q",
            dest="quiet",
            action="count",
            help="Decrease verbosity. Can be given multiple times (-qq).",
        )

        main_parser.add_argument(
            "--parser",
            choices=["python", "cpp", "legacy", "auto"],
            default="auto",
            help="Parser implementation to use (default: 'auto': use C++ parser code if available, otherwise Python).",
        )

    return main_parser


def _get_algorithm_parser() -> argparse.ArgumentParser:
    algorithm_parser = argparse.ArgumentParser(add_help=False)
    algorithm_group = algorithm_parser.add_argument_group("Generation settings")

    algorithm_group.add_argument(
        "-N",
        "--max-generations",
        type=int,
        help="Maximum number of generations to run the algorithm (ignored if --infinite is set).",
        default=fandango.api.DEFAULT_MAX_GENERATIONS,
    )
    algorithm_group.add_argument(
        "--infinite",
        action="store_true",
        help="Run the algorithm indefinitely.",
        default=False,
    )
    algorithm_group.add_argument(
        "--population-size", type=int, help="Size of the population.", default=None
    )
    algorithm_group.add_argument(
        "--elitism-rate",
        type=float,
        help="Rate of individuals preserved in the next generation.",
        default=None,
    )
    algorithm_group.add_argument(
        "--crossover-rate",
        type=float,
        help="Rate of individuals that will undergo crossover.",
        default=None,
    )
    algorithm_group.add_argument(
        "--mutation-rate",
        type=float,
        help="Rate of individuals that will undergo mutation.",
        default=None,
    )
    algorithm_group.add_argument(
        "--random-seed",
        type=int,
        help="Random seed to use for the algorithm.",
        default=None,
    )
    algorithm_group.add_argument(
        "--destruction-rate",
        type=float,
        help="Rate of individuals that will be randomly destroyed in every generation.",
        default=None,
    )
    algorithm_group.add_argument(
        "--max-repetition-rate",
        type=float,
        help="Rate at which the number of maximal repetitions should be increased.",
        default=None,
    )
    algorithm_group.add_argument(
        "--max-repetitions",
        type=int,
        help="Maximal value the number of repetitions can be increased to.",
        default=None,
    )
    algorithm_group.add_argument(
        "--max-node-rate",
        type=float,
        help="Rate at which the maximal number of nodes in a tree is increased.",
        default=None,
    )
    algorithm_group.add_argument(
        "--max-nodes",
        type=int,
        help="Maximal value, the number of nodes in a tree can be increased to.",
        default=None,
    )
    algorithm_group.add_argument(
        "-n",
        "--desired-solutions",
        "--num-outputs",
        type=int,
        help="Number of outputs to produce.",
        default=None,
    )
    algorithm_group.add_argument(
        "--best-effort",
        dest="best_effort",
        action="store_true",
        help="Produce a 'best effort' population (may not satisfy all constraints).",
        default=None,
    )
    algorithm_group.add_argument(
        "-i",
        "--initial-population",
        type=str,
        help="Directory or ZIP archive with initial population.",
        default=None,
    )
    algorithm_group.add_argument(
        "--progress-bar",
        choices=["on", "off", "auto"],
        default="auto",
        help="Whether to show the progress bar. 'auto' (default) shows the progress bar only if stderr is a terminal.",
    )

    return algorithm_parser


def _get_settings_parser(in_command_line: bool = True) -> argparse.ArgumentParser:
    settings_parser = argparse.ArgumentParser(add_help=False)
    settings_group = settings_parser.add_argument_group("General settings")

    settings_group.add_argument(
        "--warnings-are-errors",
        dest="warnings_are_errors",
        action="store_true",
        help="Treat warnings as errors.",
        default=None,
    )

    if not in_command_line:
        # Use `set -vv` or `set -q` to change logging levels
        verbosity_option = settings_group.add_mutually_exclusive_group()
        verbosity_option.add_argument(
            "--verbose",
            "-v",
            dest="verbose",
            action="count",
            help="Increase verbosity. Can be given multiple times (-vv).",
        )
        verbosity_option.add_argument(
            "--quiet",
            "-q",
            dest="quiet",
            action="store_true",
            help="Decrease verbosity. Can be given multiple times (-qq).",
        )

    return settings_parser


def _get_file_parser() -> argparse.ArgumentParser:
    file_parser = argparse.ArgumentParser(add_help=False)
    file_group = file_parser.add_argument_group("Fandango file settings")

    file_group.add_argument(
        "-f",
        "--fandango-file",
        type=argparse.FileType("r"),
        dest="fan_files",
        metavar="FAN_FILE",
        default=None,
        # required=True,
        action="append",
        help="Fandango file (.fan, .py) to be processed. Can be given multiple times. Use '-' for stdin.",
    )
    file_group.add_argument(
        "-c",
        "--constraint",
        type=str,
        dest="constraints",
        metavar="CONSTRAINT",
        default=None,
        action="append",
        help="Define an additional constraint CONSTRAINT. Can be given multiple times.",
    )
    file_group.add_argument(
        "-S",
        "--start-symbol",
        type=str,
        help="The grammar start symbol (default: '<start>').",
        default=None,
    )
    file_group.add_argument(
        "--max",
        "--maximize",
        type=str,
        dest="maxconstraints",
        metavar="MAXCONSTRAINT",
        default=None,
        action="append",
        help="Define an additional constraint MAXCONSTRAINT to be maximized. Can be given multiple times.",
    )
    file_group.add_argument(
        "--min",
        "--minimize",
        type=str,
        dest="minconstraints",
        metavar="MINCONSTRAINTS",
        default=None,
        action="append",
        help="Define an additional constraint MINCONSTRAINT to be minimized. Can be given multiple times.",
    )
    file_group.add_argument(
        "-I",
        "--include-dir",
        type=str,
        dest="includes",
        metavar="DIR",
        default=None,
        action="append",
        help="Specify a directory DIR to search for included Fandango files.",
    )
    file_group.add_argument(
        "--file-mode",
        choices=["text", "binary", "auto"],
        default="auto",
        help="Mode in which to open and write files (default is 'auto': 'binary' if grammar has bits or bytes, 'text' otherwise).",
    )
    file_group.add_argument(
        "--no-cache",
        default=True,
        dest="use_cache",
        action="store_false",
        help="Do not cache parsed Fandango files.",
    )
    file_group.add_argument(
        "--no-stdlib",
        default=True,
        dest="use_stdlib",
        action="store_false",
        help="Do not include the standard Fandango library.",
    )

    output_group = file_parser.add_argument_group("Output settings")

    output_group.add_argument(
        "-s",
        "--separator",
        type=str,
        default="\n",
        help="Output SEPARATOR between individual inputs. (default: newline).",
    )
    output_group.add_argument(
        "-d",
        "--directory",
        type=str,
        dest="directory",
        default=None,
        help="Create individual output files in DIRECTORY.",
    )
    output_group.add_argument(
        "-x",
        "--filename-extension",
        type=str,
        default=".txt",
        help="Extension of generated file names (default: '.txt').",
    )
    output_group.add_argument(
        "--format",
        choices=["string", "bits", "tree", "grammar", "value", "repr", "none"],
        default="string",
        help="Produce output(s) as string (default), as a bit string, as a derivation tree, as a grammar, as a Python value, in internal representation, or none.",
    )
    output_group.add_argument(
        "--validate",
        default=False,
        action="store_true",
        help="Run internal consistency checks for debugging.",
    )
    return file_parser


def _get_parties_parser() -> argparse.ArgumentParser:
    parties_parser = argparse.ArgumentParser(add_help=False)
    parties_group = parties_parser.add_argument_group("Party settings")
    parties_group.add_argument(
        "--party",
        action="append",
        dest="parties",
        metavar="PARTY",
        help="Only consider the PARTY part of the interaction in the .fan file.",
    )

    return parties_parser


def _populate_fuzz_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        dest="output",
        default=None,
        help="Write output to OUTPUT (default: stdout).",
    )

    command_group = parser.add_argument_group("command invocation settings")

    command_group.add_argument(
        "--input-method",
        choices=["stdin", "filename", "libfuzzer"],
        default="filename",
        help="When invoking COMMAND, choose whether Fandango input will be passed as standard input (`stdin`), as last argument on the command line (`filename`) (default), or to a libFuzzer style harness compiled to a shared .so/.dylib object (`libfuzzer`).",
    )
    command_group.add_argument(
        "test_command",
        metavar="command",
        type=str,
        nargs="?",
        help="Command to be invoked with a Fandango input.",
    )
    command_group.add_argument(
        "test_args",
        metavar="args",
        type=str,
        nargs=argparse.REMAINDER,
        help="The arguments of the command.",
    )


def _populate_parse_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "input_files",
        metavar="files",
        type=str,
        nargs="*",
        help="Files to be parsed. Use '-' for stdin.",
    )
    parser.add_argument(
        "--prefix",
        action="store_true",
        default=False,
        help="Parse a prefix only.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        dest="output",
        default=None,
        help="Write output to OUTPUT (default: none). Use '-' for stdout.",
    )


def _populate_talk_parser(parser: argparse.ArgumentParser) -> None:
    host_pattern = (
        "PORT on HOST (default: 127.0.0.1;"
        + " use '[...]' for IPv6 addresses)"
        + " using PROTOCOL ('tcp' (default)/'udp')."
    )
    parser.add_argument(
        "--client",
        metavar="[NAME=][PROTOCOL:][HOST:]PORT",
        type=str,
        help="Act as a client NAME (default: 'Client') connecting to " + host_pattern,
    )
    parser.add_argument(
        "--server",
        metavar="[NAME=][PROTOCOL:][HOST:]PORT",
        type=str,
        help="Act as a server NAME (default: 'Server') running at " + host_pattern,
    )
    parser.add_argument(
        "test_command",
        metavar="command",
        type=str,
        nargs="?",
        help="Optional command to be interacted with.",
    )
    parser.add_argument(
        "test_args",
        metavar="args",
        type=str,
        nargs=argparse.REMAINDER,
        help="The arguments of the command.",
    )


def _populate_convert_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--from",
        dest="from_format",
        choices=["antlr", "g4", "dtd", "010", "bt", "fan", "auto"],
        default="auto",
        help="Format of the external spec file: 'antlr'/'g4' (ANTLR), 'dtd' (XML DTD), '010'/'bt' (010 Editor Binary Template), 'fan' (Fandango spec), or 'auto' (default: try to guess from file extension).",
    )
    parser.add_argument(
        "--endianness", choices=["little", "big"], help="Set endianness for .bt files."
    )
    parser.add_argument(
        "--bitfield-order",
        choices=["left-to-right", "right-to-left"],
        help="Set bitfield order for .bt files.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=argparse.FileType("w"),
        dest="output",
        default=None,
        help="Write output to OUTPUT (default: stdout).",
    )
    parser.add_argument(
        "convert_files",
        type=str,
        metavar="FILENAME",
        default=None,
        nargs="+",
        help="External spec file to be converted. Use '-' for stdin.",
    )


def _populate_clear_cache_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        dest="dry_run",
        default=False,
        help="Just output the action to be performed; do not actually clear the cache.",
    )


def _populate_cd_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "directory",
        type=str,
        nargs="?",
        default=None,
        help="The directory to change into.",
    )


def _populate_shell_parser_not_in_command_line(parser: argparse.ArgumentParser) -> None:
    # Shell escape
    # Not processed by argparse,
    # but we have it here so that it is listed in help
    parser.add_argument(
        dest="shell_command",
        metavar="command",
        nargs=argparse.REMAINDER,
        default=None,
        help="The shell command to execute.",
    )


def _populate_python_parser(parser: argparse.ArgumentParser) -> None:
    # Python escape
    # Not processed by argparse,
    # but we have it here so that it is listed in help
    parser.add_argument(
        dest="python_command",
        metavar="command",
        nargs=argparse.REMAINDER,
        default=None,
        help="The Python command to execute.",
    )


def _populate_help_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "help_command",
        type=str,
        metavar="command",
        nargs="*",
        default=None,
        help="Command to get help on.",
    )



from abc import ABC, abstractmethod


class FandangoConverter(ABC):
    """Abstract superclass for converting a given grammar spec to Fandango format."""

    def __init__(self, filename: str):
        """Initialize with given grammar file"""
        self.filename = filename

    @abstractmethod
    def to_fan(self, *args, **kw_args) -> str:
        """Convert the grammar spec to Fandango format"""
        pass



# Generated from ANTLRv4Parser.g4 by ANTLR 4.13.2
from antlr4 import *

if "." in __name__:
    from .ANTLRv4Parser import ANTLRv4Parser
else:
    from ANTLRv4Parser import ANTLRv4Parser

# This class defines a complete generic visitor for a parse tree produced by ANTLRv4Parser.


class ANTLRv4ParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by ANTLRv4Parser#grammarSpec.
    def visitGrammarSpec(self, ctx: ANTLRv4Parser.GrammarSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#grammarDecl.
    def visitGrammarDecl(self, ctx: ANTLRv4Parser.GrammarDeclContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#grammarType.
    def visitGrammarType(self, ctx: ANTLRv4Parser.GrammarTypeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#prequelConstruct.
    def visitPrequelConstruct(self, ctx: ANTLRv4Parser.PrequelConstructContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#optionsSpec.
    def visitOptionsSpec(self, ctx: ANTLRv4Parser.OptionsSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#option.
    def visitOption(self, ctx: ANTLRv4Parser.OptionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#optionValue.
    def visitOptionValue(self, ctx: ANTLRv4Parser.OptionValueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#delegateGrammars.
    def visitDelegateGrammars(self, ctx: ANTLRv4Parser.DelegateGrammarsContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#delegateGrammar.
    def visitDelegateGrammar(self, ctx: ANTLRv4Parser.DelegateGrammarContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#tokensSpec.
    def visitTokensSpec(self, ctx: ANTLRv4Parser.TokensSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#channelsSpec.
    def visitChannelsSpec(self, ctx: ANTLRv4Parser.ChannelsSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#idList.
    def visitIdList(self, ctx: ANTLRv4Parser.IdListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#action_.
    def visitAction_(self, ctx: ANTLRv4Parser.Action_Context):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#actionScopeName.
    def visitActionScopeName(self, ctx: ANTLRv4Parser.ActionScopeNameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#actionBlock.
    def visitActionBlock(self, ctx: ANTLRv4Parser.ActionBlockContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#argActionBlock.
    def visitArgActionBlock(self, ctx: ANTLRv4Parser.ArgActionBlockContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#modeSpec.
    def visitModeSpec(self, ctx: ANTLRv4Parser.ModeSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#rules.
    def visitRules(self, ctx: ANTLRv4Parser.RulesContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#ruleSpec.
    def visitRuleSpec(self, ctx: ANTLRv4Parser.RuleSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#parserRuleSpec.
    def visitParserRuleSpec(self, ctx: ANTLRv4Parser.ParserRuleSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#exceptionGroup.
    def visitExceptionGroup(self, ctx: ANTLRv4Parser.ExceptionGroupContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#exceptionHandler.
    def visitExceptionHandler(self, ctx: ANTLRv4Parser.ExceptionHandlerContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#finallyClause.
    def visitFinallyClause(self, ctx: ANTLRv4Parser.FinallyClauseContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#rulePrequel.
    def visitRulePrequel(self, ctx: ANTLRv4Parser.RulePrequelContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#ruleReturns.
    def visitRuleReturns(self, ctx: ANTLRv4Parser.RuleReturnsContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#throwsSpec.
    def visitThrowsSpec(self, ctx: ANTLRv4Parser.ThrowsSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#localsSpec.
    def visitLocalsSpec(self, ctx: ANTLRv4Parser.LocalsSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#ruleAction.
    def visitRuleAction(self, ctx: ANTLRv4Parser.RuleActionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#ruleModifiers.
    def visitRuleModifiers(self, ctx: ANTLRv4Parser.RuleModifiersContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#ruleModifier.
    def visitRuleModifier(self, ctx: ANTLRv4Parser.RuleModifierContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#ruleBlock.
    def visitRuleBlock(self, ctx: ANTLRv4Parser.RuleBlockContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#ruleAltList.
    def visitRuleAltList(self, ctx: ANTLRv4Parser.RuleAltListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#labeledAlt.
    def visitLabeledAlt(self, ctx: ANTLRv4Parser.LabeledAltContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerRuleSpec.
    def visitLexerRuleSpec(self, ctx: ANTLRv4Parser.LexerRuleSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerRuleBlock.
    def visitLexerRuleBlock(self, ctx: ANTLRv4Parser.LexerRuleBlockContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerAltList.
    def visitLexerAltList(self, ctx: ANTLRv4Parser.LexerAltListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerAlt.
    def visitLexerAlt(self, ctx: ANTLRv4Parser.LexerAltContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerElements.
    def visitLexerElements(self, ctx: ANTLRv4Parser.LexerElementsContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerElement.
    def visitLexerElement(self, ctx: ANTLRv4Parser.LexerElementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerBlock.
    def visitLexerBlock(self, ctx: ANTLRv4Parser.LexerBlockContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerCommands.
    def visitLexerCommands(self, ctx: ANTLRv4Parser.LexerCommandsContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerCommand.
    def visitLexerCommand(self, ctx: ANTLRv4Parser.LexerCommandContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerCommandName.
    def visitLexerCommandName(self, ctx: ANTLRv4Parser.LexerCommandNameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerCommandExpr.
    def visitLexerCommandExpr(self, ctx: ANTLRv4Parser.LexerCommandExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#altList.
    def visitAltList(self, ctx: ANTLRv4Parser.AltListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#alternative.
    def visitAlternative(self, ctx: ANTLRv4Parser.AlternativeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#element.
    def visitElement(self, ctx: ANTLRv4Parser.ElementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#predicateOptions.
    def visitPredicateOptions(self, ctx: ANTLRv4Parser.PredicateOptionsContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#predicateOption.
    def visitPredicateOption(self, ctx: ANTLRv4Parser.PredicateOptionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#labeledElement.
    def visitLabeledElement(self, ctx: ANTLRv4Parser.LabeledElementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#ebnf.
    def visitEbnf(self, ctx: ANTLRv4Parser.EbnfContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#blockSuffix.
    def visitBlockSuffix(self, ctx: ANTLRv4Parser.BlockSuffixContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#ebnfSuffix.
    def visitEbnfSuffix(self, ctx: ANTLRv4Parser.EbnfSuffixContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#lexerAtom.
    def visitLexerAtom(self, ctx: ANTLRv4Parser.LexerAtomContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#atom.
    def visitAtom(self, ctx: ANTLRv4Parser.AtomContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#wildcard.
    def visitWildcard(self, ctx: ANTLRv4Parser.WildcardContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#notSet.
    def visitNotSet(self, ctx: ANTLRv4Parser.NotSetContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#blockSet.
    def visitBlockSet(self, ctx: ANTLRv4Parser.BlockSetContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#setElement.
    def visitSetElement(self, ctx: ANTLRv4Parser.SetElementContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#block.
    def visitBlock(self, ctx: ANTLRv4Parser.BlockContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#ruleref.
    def visitRuleref(self, ctx: ANTLRv4Parser.RulerefContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#characterRange.
    def visitCharacterRange(self, ctx: ANTLRv4Parser.CharacterRangeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#terminalDef.
    def visitTerminalDef(self, ctx: ANTLRv4Parser.TerminalDefContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#elementOptions.
    def visitElementOptions(self, ctx: ANTLRv4Parser.ElementOptionsContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#elementOption.
    def visitElementOption(self, ctx: ANTLRv4Parser.ElementOptionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#identifier.
    def visitIdentifier(self, ctx: ANTLRv4Parser.IdentifierContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ANTLRv4Parser#qualifiedIdentifier.
    def visitQualifiedIdentifier(self, ctx: ANTLRv4Parser.QualifiedIdentifierContext):
        return self.visitChildren(ctx)


del ANTLRv4Parser



from typing import Any, Optional

from fandango.language.tree import DerivationTree
from fandango.language.symbols import NonTerminal, Symbol


class ParseState:
    def __init__(
        self,
        nonterminal: NonTerminal,
        position: int,
        symbols: tuple[tuple[Symbol, frozenset[tuple[str, Any]]], ...],
        dot: int = 0,
        children: Optional[list[DerivationTree]] = None,
        is_incomplete: bool = False,
        incomplete_idx=0,
    ):
        self._nonterminal = nonterminal
        self._position = position
        self._symbols = symbols
        self._dot = dot
        self.children = children or []
        self.is_incomplete = is_incomplete
        self.incomplete_idx = incomplete_idx
        self._hash: Optional[int] = None

    @property
    def nonterminal(self):
        return self._nonterminal

    def append_child(self, child: DerivationTree):
        self.children.append(child)
        self._hash = None

    def extend_children(self, children: list[DerivationTree]):
        self.children.extend(children)
        self._hash = None

    @property
    def position(self):
        return self._position

    @property
    def symbols(self):
        return self._symbols

    @property
    def dot(self) -> Optional[Symbol]:
        return self.symbols[self._dot][0] if self._dot < len(self.symbols) else None

    @property
    def dot_params(self) -> Optional[frozenset[tuple[str, Any]]]:
        return self.symbols[self._dot][1] if self._dot < len(self.symbols) else None

    def finished(self):
        return self._dot >= len(self.symbols) and not self.is_incomplete

    def next_symbol_is_nonterminal(self):
        return (
            self._dot < len(self.symbols) and self.symbols[self._dot][0].is_non_terminal
        )

    def next_symbol_is_terminal(self):
        return self._dot < len(self.symbols) and self.symbols[self._dot][0].is_terminal

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(
                (
                    self.nonterminal,
                    self.position,
                    self.symbols,
                    self._dot,
                    tuple(self.children),
                )
            )
        return self._hash

    def __eq__(self, other):
        return (
            isinstance(other, ParseState)
            and self.nonterminal == other.nonterminal
            and self.position == other.position
            and self.symbols == other.symbols
            and self._dot == other._dot
        )

    def __repr__(self):
        return (
            f"({self.nonterminal.format_as_spec()} -> "
            + "".join(
                [
                    f"{'•' if i == self._dot else ''}{s[0].format_as_spec()!s}"
                    for i, s in enumerate(self.symbols)
                ]
            )
            + ("•" if self.finished() else "")
            + f", column {self.position}"
            + ")"
        )

    def next(self):
        next_state = self.copy()
        next_state._dot += 1
        return next_state

    def copy(self):
        return ParseState(
            self.nonterminal,
            self.position,
            self.symbols,
            self._dot,
            self.children[:],
            self.is_incomplete,
            self.incomplete_idx,
        )



import abc
import enum
from typing import Optional, Any

from fandango.constraints.base import GeneticBase
from fandango.constraints.failing_tree import FailingTree
from fandango.language.tree import DerivationTree


class Fitness(abc.ABC):
    """
    Abstract class to represent the fitness of a tree.
    """

    def __init__(
        self, success: bool, failing_trees: Optional[list[FailingTree]] = None
    ):
        """
        Initialize the Fitness with the given success and failing trees.

        :param bool success: The success of the fitness.
        :param Optional[list[FailingTree]] failing_trees: The list of failing trees.
        """
        self.success = success
        self.failing_trees = failing_trees or []

    @abc.abstractmethod
    def fitness(self) -> float:
        """
        Abstract method to calculate the fitness of the tree.
        :return float: The fitness of the tree.
        """
        pass

    @abc.abstractmethod
    def __copy__(self) -> "Fitness":
        pass

    def __repr__(self):
        return f"Fitness(success={self.success})"


class ValueFitness(Fitness):
    """
    Class to represent the fitness of a tree based on calculated values.
    The fitness is calculated as the average of the values.
    This class contrast the `ConstraintFitness` class, which calculates the fitness based on the number of
    constraints satisfied.
    """

    def __init__(
        self,
        values: Optional[list[float]] = None,
        failing_trees: Optional[list[FailingTree]] = None,
    ):
        """
        Initialize the ValueFitness with the given values and failing trees.
        :param Optional[list[float]] values: The list of values.
        :param Optional[list[FailingTree]] failing_trees: The list of failing trees.
        """
        super().__init__(True, failing_trees)
        self.values = values or []

    def fitness(self) -> float:
        """
        Calculate the fitness of the tree as the average of the values.
        :return float: The fitness of the tree.
        """
        if self.values:
            try:
                return sum(self.values) / len(self.values)
            except OverflowError:
                # OverflowError: integer division result too large for a float
                return sum(self.values) // len(self.values)
        else:
            return 0

    def __copy__(self) -> Fitness:
        return ValueFitness(self.values[:])

    def __repr__(self):
        return f"ValueFitness(values={self.values})"


class ConstraintFitness(Fitness):
    """
    Class to represent the fitness of a tree based on constraints.
    The fitness is calculated as the number of constraints solved by the tree divided by the total number of
    constraints.
    """

    def __init__(
        self,
        solved: int,
        total: int,
        success: bool,
        failing_trees: Optional[list[FailingTree]] = None,
    ):
        """
        Initialize the ConstraintFitness with the given solved, total, success, and failing trees.
        :param int solved: The number of constraints solved by the tree.
        :param int total: The total number of constraints.
        :param bool success: The success of the fitness.
        :param Optional[list[FailingTree]] failing_trees: The list of failing trees.
        """
        super().__init__(success, failing_trees)
        self.solved = solved
        self.total = total

    def fitness(self) -> float:
        """
        Calculate the fitness of the tree as the number of constraints solved by the tree divided by the total number of
        constraints.
        :return float: The fitness of the tree.
        """
        if self.total:
            return self.solved / self.total
        else:
            return 0

    def __copy__(self) -> Fitness:
        return ConstraintFitness(
            solved=self.solved,
            total=self.total,
            success=self.success,
            failing_trees=self.failing_trees[:],
        )

    def __repr__(self):
        return f"ConstraintFitness(solved={self.solved}, total={self.total}, success={self.success})"



from copy import copy
import enum
from typing import Any, Optional
from fandango.constraints.failing_tree import Comparison, ComparisonSide
from fandango.language.tree import DerivationTree
from fandango.constraints.constraint_visitor import ConstraintVisitor
from fandango.constraints.constraint import Constraint
from fandango.constraints.fitness import (
    ConstraintFitness,
    FailingTree,
)
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.logger import LOGGER, print_exception


class ComparisonConstraint(Constraint):
    """
    Represents a comparison constraint that can be used for fitness evaluation.
    """

    def __init__(self, operator: Comparison, left: str, right: str, *args, **kwargs):
        """
        Initializes the comparison constraint with the given operator, left side, and right side.
        :param Comparison operator: The operator to use.
        :param str left: The left side of the comparison.
        :param str right: The right side of the comparison.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.operator = operator
        self.left = left
        self.right = right
        self.types_checked = False

    def fitness(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> ConstraintFitness:
        """
        Calculate the fitness of the tree based on the given comparison.
        """
        tree_hash = self.get_hash(tree, scope, population, local_variables)
        # If the fitness has already been calculated, return the cached value
        if tree_hash in self.cache:
            return copy(self.cache[tree_hash])
        # Initialize the fitness values
        solved = 0
        total = 0
        failing_trees = []
        has_combinations = False
        # If the tree is None, the fitness is 0
        if tree is None:
            return ConstraintFitness(0, 0, False)
        # Iterate over all combinations of the tree and the scope
        for combination in self.combinations(tree, scope, population):
            total += 1
            has_combinations = True
            # Update the local variables to initialize the placeholders with the values of the combination
            local_vars = self.local_variables.copy()
            if local_variables:
                local_vars.update(local_variables)
            local_vars.update(
                {name: container.evaluate() for name, container in combination}
            )
            # Evaluate the left and right side of the comparison
            try:
                left = self.eval(self.left, self.global_variables, local_vars)
            except Exception as e:
                print_exception(e, f"Evaluation failed: {self.left}")
                continue

            try:
                right = self.eval(self.right, self.global_variables, local_vars)
            except Exception as e:
                print_exception(e, f"Evaluation failed: {self.right}")
                continue

            if not hasattr(self, "types_checked") or not self.types_checked:
                self.types_checked = self.check_type_compatibility(left, right)

            # Initialize the suggestions
            suggestions = []
            is_solved = False
            match self.operator:
                case Comparison.EQUAL:
                    # If the left and right side are equal, the constraint is solved
                    if left == right:
                        is_solved = True
                    else:
                        # If the left and right side are not equal, add suggestions to the list
                        if not self.right.strip().startswith("len("):
                            suggestions.append(
                                (Comparison.EQUAL, left, ComparisonSide.RIGHT)
                            )
                        if not self.left.strip().startswith("len("):
                            suggestions.append(
                                (Comparison.EQUAL, right, ComparisonSide.LEFT)
                            )
                case Comparison.NOT_EQUAL:
                    # If the left and right side are not equal, the constraint is solved
                    if left != right:
                        is_solved = True
                    else:
                        # If the left and right side are equal, add suggestions to the list
                        suggestions.append(
                            (Comparison.NOT_EQUAL, left, ComparisonSide.RIGHT)
                        )
                        suggestions.append(
                            (Comparison.NOT_EQUAL, right, ComparisonSide.LEFT)
                        )
                case Comparison.GREATER:
                    # If the left side is greater than the right side, the constraint is solved
                    if left > right:
                        is_solved = True
                    else:
                        # If the left side is not greater than the right side, add suggestions to the list
                        suggestions.append(
                            (Comparison.LESS, left, ComparisonSide.RIGHT)
                        )
                        suggestions.append(
                            (Comparison.GREATER, right, ComparisonSide.LEFT)
                        )
                case Comparison.GREATER_EQUAL:
                    # If the left side is greater than or equal to the right side, the constraint is solved
                    if left >= right:
                        is_solved = True
                    else:
                        # If the left side is not greater than or equal to the right side, add suggestions to the list
                        suggestions.append(
                            (Comparison.LESS_EQUAL, left, ComparisonSide.RIGHT)
                        )
                        suggestions.append(
                            (Comparison.GREATER_EQUAL, right, ComparisonSide.LEFT)
                        )
                case Comparison.LESS:
                    # If the left side is less than the right side, the constraint is solved
                    if left < right:
                        is_solved = True
                    else:
                        # If the left side is not less than the right side, add suggestions to the list
                        suggestions.append(
                            (Comparison.GREATER, left, ComparisonSide.RIGHT)
                        )
                        suggestions.append(
                            (Comparison.LESS, right, ComparisonSide.LEFT)
                        )
                case Comparison.LESS_EQUAL:
                    # If the left side is less than or equal to the right side, the constraint is solved
                    if left <= right:
                        is_solved = True
                    else:
                        # If the left side is not less than or equal to the right side, add suggestions to the list
                        suggestions.append(
                            (Comparison.GREATER_EQUAL, left, ComparisonSide.RIGHT)
                        )
                        suggestions.append(
                            (Comparison.LESS_EQUAL, right, ComparisonSide.LEFT)
                        )
            if is_solved:
                solved += 1
            else:
                # If the comparison is not solved, add the failing trees to the list
                for _, container in combination:
                    for node in container.get_trees():
                        ft = FailingTree(node, self, suggestions=suggestions)
                        # if ft not in failing_trees:
                        # failing_trees.append(ft)
                        failing_trees.append(ft)

        if not has_combinations:
            solved += 1
            total += 1

        # Create the fitness object
        fitness = ConstraintFitness(
            solved, total, solved == total, failing_trees=failing_trees
        )
        # Cache the fitness
        self.cache[tree_hash] = fitness
        return fitness

    def check_type_compatibility(self, left: Any, right: Any) -> bool:
        """
        Check the types of `left` and `right` are compatible in a comparison.
        Return True iff check was actually performed
        """
        if left is None and right is None:
            return True

        if left is None or right is None:
            # Cannot check - value does not exist
            return False

        if isinstance(left, type(right)):
            return True

        if isinstance(left, DerivationTree):
            return left.value().can_compare_with(right)
        if isinstance(right, DerivationTree):
            return right.value().can_compare_with(left)

        if isinstance(left, (bool, int, float)) and isinstance(
            right, (bool, int, float)
        ):
            return True

        LOGGER.warning(
            f"{self.format_as_spec()}: {self.operator.value!r}: Cannot compare {type(left).__name__!r} and {type(right).__name__!r}"
        )
        return True

    def format_as_spec(self) -> str:
        representation = f"{self.left} {self.operator.value} {self.right}"
        for identifier in self.searches:
            representation = representation.replace(
                identifier, self.searches[identifier].format_as_spec()
            )
        return representation

    def accept(self, visitor: "ConstraintVisitor"):
        """
        Accepts a visitor to traverse the constraint structure.
        :param ConstraintVisitor visitor: The visitor to accept.
        """
        return visitor.visit_comparison_constraint(self)



from __future__ import annotations
from types import NoneType
from typing import Any, Optional
import warnings
import enum

from fandango.errors import FandangoConversionError, FandangoValueError

STRING_TO_BYTES_ENCODING = "utf-8"  # according to the docs
BYTES_TO_STRING_ENCODING = "latin-1"  # according to the docs


class TreeValueType(enum.Enum):
    STRING = "string"
    BYTES = "bytes"
    TRAILING_BITS_ONLY = "trailing_bits"
    EMPTY = "empty"


def trailing_bits_to_int(trailing_bits: list[int]) -> int:
    """
    Convert a list of trailing bits to an integer.

    The bits are assumed to be in big endian order and are not checked for validity (assumed to be 0 or 1)
    """
    return sum(bit << i for i, bit in enumerate(reversed(trailing_bits)))


def _str_to_bytes(value: str, encoding: str) -> bytes:
    """
    Convert a string to bytes.

    :param value: The string to convert.
    :param encoding: The encoding to use when converting the string to bytes.
    :throws FandangoConversionError: If the string cannot be converted to bytes.
    :return: A bytes object.
    """
    try:
        return value.encode(encoding=encoding)
    except UnicodeEncodeError as e:
        raise FandangoConversionError(
            f"string to bytes conversion failed, string: {value}, encoding: {encoding}, error: {e}"
        )


def _bytes_to_str(value: bytes, encoding: str) -> str:
    """
    Convert bytes to a string.

    :param value: The bytes to convert.
    :param encoding: The encoding to use when converting the bytes to a string.
    :throws FandangoConversionError: If the bytes cannot be converted to a string.
    :return: A string.
    """
    try:
        return value.decode(encoding=encoding)
    except UnicodeDecodeError as e:
        raise FandangoConversionError(
            f"bytes to string conversion failed, bytes: {value!r}, encoding: {encoding}, error: {e}"
        )


def _get_exclusive_base_type_method_is_implemented_for(name: str) -> Optional[type]:
    """
    If the method, whose name is passed, is implemented on only one of int, str, bytes, return that type.
    If the method is implemented on more than one type, return None.
    If the method is not implemented on any type, raise an error.
    """
    is_in_int = name in dir(1)
    is_in_str = name in dir("1")
    is_in_bytes = name in dir(b"1")
    match is_in_int + is_in_str + is_in_bytes:
        case 0:
            raise FandangoValueError(
                f"Method {name} is not implemented on any underlying type"
            )
        case 1:
            return int if is_in_int else str if is_in_str else bytes
        case _:
            return None


def _attach_to_first_arg(
    function_names: list[str],
):
    """
    Decorator to add methods to a class, where self is transformed to either
    (a) the only base type it is implemented on,
    (b) the type of the first argument,
    (c) the type of the only keyword argument.

    If the first argument is a DerivationTree or TreeValue, it is converted to the underlying type. This is deprecated.
    """

    def make_method(name):
        def method(self, *args, **kwargs):
            base_type = _get_exclusive_base_type_method_is_implemented_for(name)
            if base_type is None:
                message = (
                    "Using a {} arg to pass as a {} arg "
                    f"to {name} is deprecated "
                    "because there is no automated way of knowing which type to transform the underlying type to. "
                    "Use `str` or `bytes` to explicitly convert to the desired type first."
                )

                if len(args) == 0:
                    assert (
                        len(kwargs) == 1
                    ), f"Method {name} must have at least one unnamed argument or exactly one named argument"
                    k, v = list(kwargs.items())[0]
                    # cannot import DerivationTree because of circular import
                    if v.__class__.__name__ == "DerivationTree":
                        v = v.value()._inner_value(
                            message.format("DerivationTree", "keyword")
                        )
                    elif isinstance(v, TreeValue):
                        v = v._inner_value(message.format("TreeValue", "keyword"))
                    base_type = type(v)
                    kwargs = {k: v}
                else:
                    first_arg = args[0]
                    # cannot import DerivationTree because of circular import
                    if first_arg.__class__.__name__ == "DerivationTree":
                        first_arg = first_arg.value()._inner_value(
                            message.format("DerivationTree", "first")
                        )
                    elif isinstance(first_arg, TreeValue):
                        first_arg = first_arg._inner_value(
                            message.format("TreeValue", "first")
                        )
                    base_type = type(first_arg)
                    args = (first_arg,) + args[1:]

            assert base_type in [
                str,
                bytes,
                int,
            ], f"Cannot determine the type the base should be converted to based on argument of type {base_type.__name__}"

            base = base_type(self)
            return getattr(base, name)(*args, **kwargs)

        return method

    def decorator(cls):
        for name in function_names:
            # Check if the method is actually implemented on the class itself, not inherited
            if name in cls.__dict__:
                warnings.warn(
                    f"Method {name} already exists on {cls.__name__}, skipping",
                    Warning,
                )
            else:
                setattr(cls, name, make_method(name))
        return cls

    return decorator


def _attach_to_underlying(
    function_names: list[str],
):
    """
    Decorator to add methods to a class, where self is transformed to either (a) the only base type it is implemented on or (b) the underlying type and then the method is called on that.

    (b) is deprecated.
    """

    def make_method(name):
        def method(self: TreeValue, *args, **kwargs):
            base_type = _get_exclusive_base_type_method_is_implemented_for(name)
            if base_type is not None:
                base = base_type(self)
            else:
                message = (
                    f"Using {name} on DerivationTree or TreeValue objects is deprecated "
                    "because there is no automated way of knowing which type to transform the underlying type to. "
                    "Use `str` or `bytes` to explicitly convert to the desired type first."
                )
                base = self._inner_value(message)
            return getattr(base, name)(*args, **kwargs)

        return method

    def decorator(cls):
        for name in function_names:
            # Check if the method is actually implemented on the class itself, not inherited
            if name in cls.__dict__:
                warnings.warn(
                    f"Method {name} already exists on {cls.__name__}, skipping",
                    Warning,
                )
            else:
                setattr(cls, name, make_method(name))
        return cls

    return decorator


# not implemented on purpose:
# __bool__
# __buffer__
# __bytes__
# __delattr__
# __dir__
# __doc__
# __eq__
# __format__
# __getattribute__
# __getitem__
# __getnewargs__
# __getstate__
# __hash__
# __init__
# __init_subclass__
# __int__
# __iter__
# __len__
# __ne__
# __new__
# __reduce__
# __reduce_ex__
# __repr__
# __setattr__
# __sizeof__
# __str__
# __subclasshook__
# decode
# denominator
# encode
# fromhex
# from_bytes
# imag
# numerator
# maketrans
# join
# real
# replace
# translate
# to_bytes

DIRECT_ACCESS_METHODS_BASE_TO_FIRST_ARG_TYPE = [
    "__add__",
    "__contains__",
    "__divmod__",
    "__floordiv__",
    "__ge__",
    "__gt__",
    "__le__",
    "__lshift__",
    "__lt__",
    "__mod__",
    "__radd__",
    "__rand__",
    "__rdivmod__",
    "__rfloordiv__",
    "__rlshift__",
    "__rmod__",
    "__rrshift__",
    "__rshift__",
    "__rsub__",
    "__rtruediv__",
    "__rxor__",
    "__sub__",
    "__truediv__",
    "__xor__",
    "count",
    "endswith",
    "find",
    "index",
    "partition",
    "removeprefix",
    "removesuffix",
    "rfind",
    "rindex",
    "rpartition",
    "startswith",
]

DIRECT_ACCESS_METHODS_BASE_TO_UNDERLYING_TYPE = [
    "__and__",
    "__abs__",
    "__ceil__",
    "__float__",
    "__floor__",
    "__index__",
    "__invert__",
    "__mul__",
    "__neg__",
    "__or__",
    "__pos__",
    "__pow__",
    "__rmul__",
    "__round__",
    "__ror__",
    "__rpow__",
    "__trunc__",
    "as_integer_ratio",
    "bit_count",
    "bit_length",
    "capitalize",
    "casefold",
    "center",
    "conjugate",
    "expandtabs",
    "format",
    "format_map",
    "hex",
    "isalnum",
    "isalpha",
    "isascii",
    "isdecimal",
    "isdigit",
    "isidentifier",
    "islower",
    "isnumeric",
    "isprintable",
    "isspace",
    "istitle",
    "isupper",
    "lower",
    "lstrip",
    "ljust",
    "split",
    "strip",
    "rjust",
    "rsplit",
    "rstrip",
    "splitlines",
    "swapcase",
    "title",
    "upper",
    "zfill",
]


@_attach_to_first_arg(DIRECT_ACCESS_METHODS_BASE_TO_FIRST_ARG_TYPE)
@_attach_to_underlying(DIRECT_ACCESS_METHODS_BASE_TO_UNDERLYING_TYPE)
class TreeValue:
    def __init__(
        self,
        value: str | bytes | int | None,
        *,
        trailing_bits: list[int] = [],
        allow_empty: bool = False,
    ):
        self._value: str | bytes | None
        assert all(
            bit & 1 == bit for bit in trailing_bits
        ), "trailing bits must be 0 or 1, got " + str(trailing_bits)

        if isinstance(value, int):
            assert (
                value & 1 == value
            ), "ints are used for bit values, and must thus be 0 or 1"
            assert trailing_bits == [], "trailing bits are not supported for int values"
            trailing_bits = [value]
            value = None

        if value is None:
            assert (
                allow_empty or len(trailing_bits) > 0
            ), "None values must have trailing bits"

        self._value = value
        self._trailing_bits = trailing_bits

    def _reduce_trailing_bits(self, str_to_bytes_encoding: str) -> None:
        """
        Reduce the trailing bits into the value.

        :raises FandangoConversionError: If the trailing bits don't sum up to a perfect number of bytes.
        """
        if not self._trailing_bits:
            return

        if len(self._trailing_bits) % 8 != 0:
            raise FandangoConversionError(
                "Trailing bits are not a multiple of 8, currently have "
                f"{len(self._trailing_bits)} bits"
            )

        num = trailing_bits_to_int(self._trailing_bits)
        bytes_ = num.to_bytes(len(self._trailing_bits) // 8)
        self._trailing_bits = []
        if isinstance(self._value, str):
            self._value = (
                _str_to_bytes(self._value, encoding=str_to_bytes_encoding) + bytes_
            )
        elif isinstance(self._value, bytes):
            self._value = self._value + bytes_
        elif self._value is None:
            self._value = bytes_
        else:
            raise FandangoValueError(
                f"Invalid value type: {type(self._value)}, {self._trailing_bits}. This should not happen, please report this as a bug"
            )

    @property
    def type_(self) -> TreeValueType:
        if self._value is None:
            if self._trailing_bits:
                return TreeValueType.TRAILING_BITS_ONLY
            else:
                return TreeValueType.EMPTY
        elif isinstance(self._value, str):
            if self._trailing_bits:
                # Trailing bits are reduced to bytes
                return TreeValueType.BYTES
            else:
                return TreeValueType.STRING
        elif isinstance(self._value, bytes):
            return TreeValueType.BYTES
        else:
            raise ValueError(f"Invalid value type: {type(self._value)}")

    def is_type(self, type_: TreeValueType) -> bool:
        return self.type_ == type_

    def count_bytes(self) -> int:
        """
        Count the number of bytes in the TreeValue.

        :throws FandangoConversionError: If the value cannot be converted to bytes.
        :return: The number of bytes in the TreeValue.
        """
        return len(self.to_bytes())

    def append(
        self,
        other: TreeValue,
        str_to_bytes_encoding: str = STRING_TO_BYTES_ENCODING,
    ) -> TreeValue:
        """
        Create a new TreeValue by appending another value to this value.

        :param other: The value to append..
        :param str_to_bytes_encoding: The encoding to use when converting strings to bytes.
        :throws FandangoConversionError: If one of the values has to be converted and the conversion fails.
        :return: A new TreeValue.
        """

        if self.is_type(TreeValueType.EMPTY):
            return TreeValue(other._value, trailing_bits=other._trailing_bits)

        if other._value is None:
            trailing_bits = self._trailing_bits + other._trailing_bits
            return TreeValue(self._value, trailing_bits=trailing_bits)

        # flush bits, will set self._value
        self._reduce_trailing_bits(str_to_bytes_encoding=str_to_bytes_encoding)

        if isinstance(self._value, str):
            if isinstance(other._value, str):
                return TreeValue(
                    self._value + other._value,
                    trailing_bits=other._trailing_bits,
                )
            elif isinstance(other._value, bytes):
                return TreeValue(
                    _str_to_bytes(self._value, encoding=str_to_bytes_encoding)
                    + other._value,
                    trailing_bits=other._trailing_bits,
                )

        elif isinstance(self._value, bytes):
            if isinstance(other._value, str):
                return TreeValue(
                    self._value
                    + _str_to_bytes(other._value, encoding=str_to_bytes_encoding),
                    trailing_bits=other._trailing_bits,
                )
            elif isinstance(other._value, bytes):
                return TreeValue(
                    self._value + other._value,
                    trailing_bits=other._trailing_bits,
                )
        raise FandangoValueError(
            f"Cannot compute {self._value!r} + {other._value!r}. This should not happen, please report this as a bug"
        )

    def can_compare_with(self, right: object) -> bool:
        if isinstance(right, TreeValue):
            return self.is_type(right.type_)

        return (
            (isinstance(right, int) and self.is_type(TreeValueType.TRAILING_BITS_ONLY))
            or (isinstance(right, str) and self.is_type(TreeValueType.STRING))
            or (isinstance(right, bytes) and self.is_type(TreeValueType.BYTES))
        )

    def to_string(
        self,
        bytes_to_str_encoding: str = BYTES_TO_STRING_ENCODING,
    ) -> str:
        """
        Convert the TreeValue to a string.

        :param bytes_to_str_encoding: The encoding to use when converting bytes to strings.
        :throws FandangoConversionError: If the value cannot be converted to a string.
        :return: A string.
        """
        if self.is_type(TreeValueType.EMPTY):
            return ""

        # encoding with the same encoding in both direction
        self._reduce_trailing_bits(str_to_bytes_encoding=bytes_to_str_encoding)
        if isinstance(self._value, str):
            return self._value
        if isinstance(self._value, bytes):
            return _bytes_to_str(self._value, encoding=bytes_to_str_encoding)
        raise FandangoValueError(
            f"Invalid value type: {type(self._value)}, {self._trailing_bits}. This should not happen, please report this as a bug"
        )

    def to_bytes(
        self,
        str_to_bytes_encoding: str = STRING_TO_BYTES_ENCODING,
    ) -> bytes:
        """
        Convert the TreeValue to bytes.

        :param str_to_bytes_encoding: The encoding to use when converting strings to bytes.
        :throws FandangoConversionError: If the value cannot be converted to bytes.
        :return: A bytes object.
        """
        if self.is_type(TreeValueType.EMPTY):
            return b""

        self._reduce_trailing_bits(str_to_bytes_encoding=str_to_bytes_encoding)
        if isinstance(self._value, bytes):
            return self._value
        elif isinstance(self._value, str):
            return _str_to_bytes(self._value, encoding=str_to_bytes_encoding)
        raise FandangoValueError(
            f"Invalid value type: {type(self._value)}, {self._trailing_bits}. This should not happen, please report this as a bug"
        )

    def to_bits(
        self,
        str_to_bytes_encoding: str = STRING_TO_BYTES_ENCODING,
    ) -> str:
        """
        Convert the TreeValue to a string of 0s and 1s representing bits.

        :param str_to_bytes_encoding: The encoding to use when converting strings to bytes.
        :throws FandangoConversionError: If the value cannot be converted to bits.
        :return: A string of bits.
        """
        if self._value is None:
            value = ""
        elif isinstance(self._value, bytes):
            value = "".join(f"{byte_:08b}" for byte_ in self._value)
        elif isinstance(self._value, str):
            value = "".join(
                f"{byte_:08b}"
                for byte_ in _str_to_bytes(self._value, encoding=str_to_bytes_encoding)
            )
        else:
            raise FandangoValueError(
                f"Invalid value type: {type(self._value)}. This should not happen, please report this as a bug"
            )

        trailing_bits = "".join(str(bit) for bit in self._trailing_bits)

        return value + trailing_bits

    def to_int(
        self,
        str_to_bytes_encoding: str = STRING_TO_BYTES_ENCODING,
        bytes_to_str_encoding: str = BYTES_TO_STRING_ENCODING,
    ) -> int:
        """
        Convert the TreeValue to an integer.

        Strings are converted to ints using int(), bytes are converted to strings using the given encoding and then to ints using int().

        If bits were ever included, the underlying data structure is bytes.

        :param str_to_bytes_encoding: The encoding to use when converting strings to bytes.
        :throws FandangoConversionError: If the value cannot be converted to an integer.
        :return: An integer.
        """
        if self._value is None:  # only trailing bits
            return trailing_bits_to_int(self._trailing_bits)
        else:
            self._reduce_trailing_bits(str_to_bytes_encoding=str_to_bytes_encoding)
            if isinstance(self._value, str):
                return int(self._value)
            elif isinstance(self._value, bytes):
                try:
                    return int(
                        self.to_string(bytes_to_str_encoding=bytes_to_str_encoding)
                    )
                except ValueError as e:
                    raise FandangoConversionError(
                        f"int conversion failed, value: {self._value!r}, encoding: {bytes_to_str_encoding}, error: {e}"
                    )
            else:
                raise FandangoValueError(
                    f"Invalid value type: {type(self._value)}, {self._trailing_bits}. This should not happen, please report this as a bug"
                )

    @property
    def trailing_bits(self) -> Optional[tuple[int, ...]]:
        """
        Return the trailing bits as a tuple of 0s and 1s. MSB first. None if this is not only trailing bits.

        :return: A tuple of integers, or None if this is not only trailing bits.
        """
        if self._value is None:
            return tuple(self._trailing_bits)
        else:
            return None

    def __str__(self) -> str:
        return self.to_string()

    def __bytes__(self) -> bytes:
        return self.to_bytes()

    def __int__(self) -> int:
        return self.to_int()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, TreeValue):
            return (
                self._value == other._value
                and self._trailing_bits == other._trailing_bits
            )
        # attempt to coerce type of self to the same as other
        # works (at least) for int, str, bytes
        if isinstance(other, int):
            return self.to_int() == other
        elif isinstance(other, str):
            return self.to_string() == other
        elif isinstance(other, bytes):
            return self.to_bytes() == other
        else:
            raise TypeError(
                f"Cannot compare TreeValue of type {type(self)} with {type(other)}"
            )

    def __hash__(self) -> int:
        return hash((self._value, tuple(self._trailing_bits)))

    def __repr__(self) -> str:
        if self._trailing_bits:
            return f"{self._value!r} + bits: {''.join(str(bit) for bit in self._trailing_bits)}"
        else:
            return repr(self._value)

    def __deepcopy__(self, memo: dict[int, Any]) -> TreeValue:
        return TreeValue(
            self._value, trailing_bits=self._trailing_bits, allow_empty=True
        )

    def _inner_value(self, warn_message: str) -> str | bytes | int:
        """
        Don't you dare using this outside this class! This is a dirty, dirty hack to support the legacy interface.
        """

        warnings.warn(warn_message, DeprecationWarning, stacklevel=2)

        match self.type_:
            case TreeValueType.STRING:
                return str(self)
            case TreeValueType.BYTES:
                return bytes(self)
            case TreeValueType.TRAILING_BITS_ONLY:
                return int(self)
            case TreeValueType.EMPTY:
                return ""
            case _:
                raise FandangoValueError(
                    f"Invalid value type: {type(self)}. This should not happen, please report this as a bug"
                )

    @classmethod
    def empty(cls) -> TreeValue:
        return TreeValue(None, trailing_bits=[], allow_empty=True)



#!/usr/bin/env pytest

import itertools
import random
import unittest
import logging

from fandango import Fandango, FandangoParseError
from .utils import DOCS_ROOT


class APITest(unittest.TestCase):
    SPEC_abc = r"""
    <start> ::= ('a' | 'b' | 'c')+
    where str(<start>) != 'd'
    """

    SPEC_abcd = r"""
    <start> ::= ('a' | 'b' | 'c')+ 'd'
    where str(<start>) != 'd'
    """

    def test_fuzz(self):
        with open(DOCS_ROOT / "persons-faker.fan") as persons:
            fan = Fandango(persons)

        random.seed(0)
        for tree in itertools.islice(fan.generate_solutions(), 10):
            print(str(tree))

    def test_fuzz_from_string(self):
        fan = Fandango(self.SPEC_abc, logging_level=logging.INFO)
        random.seed(0)
        for tree in itertools.islice(fan.generate_solutions(), 10):
            print(str(tree))

    def test_parse(self):
        fan = Fandango(self.SPEC_abc)
        word = "abc"

        for tree in fan.parse(word):
            assert tree is not None
            print(f"tree = {repr(str(tree))}")
            print(tree.to_grammar())

    def test_incomplete_parse(self):
        fan = Fandango(self.SPEC_abcd)
        word = "ab"

        for tree in fan.parse(word, prefix=True):
            assert tree is not None
            print(f"tree = {repr(str(tree))}")
            print(tree.to_grammar())

    def test_failing_incomplete_parse(self):
        fan = Fandango(self.SPEC_abcd)
        invalid_word = "ab"

        with self.assertRaises(FandangoParseError):
            list(fan.parse(invalid_word))  # force generator evaluation

    def test_failing_parse(self):
        fan = Fandango(self.SPEC_abcd)
        invalid_word = "abcdef"

        with self.assertRaises(FandangoParseError):
            list(fan.parse(invalid_word))  # force generator evaluation

    def ensure_capped_generation(self):
        fan = Fandango(self.SPEC_abcd, logging_level=logging.INFO)
        solutions = fan.fuzz()
        self.assertLess(
            100,
            len(solutions),
            f"Expected more than 100 trees, only received {len(solutions)}",
        )


if __name__ == "__main__":
    unittest.main()



import copy
import re
from typing import Optional

from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.nodes.node import Node, NodeSettings
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode


class GrammarSetting(HasSettings):
    ALL_WITH_TYPE_RE = re.compile(r"all_with_type\((.*?)\)")

    def __init__(self, selector: str, rules: dict[str, str]):

        self._selector = selector
        self._node_settings = NodeSettings(rules)

    def _matches(self, node: Node) -> bool:
        if self._selector == "*":
            return True
        if match := self.ALL_WITH_TYPE_RE.match(self._selector):
            if type(node).__name__ == match.group(1):
                return True
        if isinstance(node, NonTerminalNode):
            if node.symbol.name() == self._selector:
                return True
        return False

    def settings_for(self, node: "Node") -> Optional[NodeSettings]:
        if not self._matches(node):
            return None
        return copy.deepcopy(self._node_settings)
